# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import re
from typing import Annotated
from typing import Literal

from aiopki.types import EncryptionResult
from pydantic.functional_validators import AfterValidator

from .baseclaim import BaseClaim


__all__: list[str] = ['BirthdateClaim']


def validate_year(value: str) -> str:
    if not re.match('^[0-9]{4}$', value):
        raise ValueError("Invalid birthdate format.")
    return value


Birthyear = Annotated[str, AfterValidator(validate_year)]
BirthdateType = datetime.datetime | datetime.date | Birthyear | Literal['0000']


class BirthdateClaim(BaseClaim):
    kind: Literal['birthdate'] = 'birthdate'
    value: BirthdateType | EncryptionResult

    def get_encryption_input(self) -> bytes:
        return self.get_masking_input()

    @BaseClaim.requires_decryption
    def get_masking_input(self) -> bytes:
        assert not isinstance(self.value, EncryptionResult)
        if isinstance(self.value, (datetime.datetime, datetime.date)):
            v = self.value.isoformat()
        else:
            v = str(self.value)
        return str.encode(v, 'ascii')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any
from typing import Callable
from typing import TypeVar

import pydantic

from oauthx.lib.exceptions import TrustIssues
from .authorizationrequestparameters import AuthorizationRequestParameters
from .clientauthorizationstatekey import ClientAuthorizationStateKey
from .redirectionparameters import RedirectionParameters


T = TypeVar('T')


class ClientAuthorizationState(pydantic.BaseModel):
    """Maintains state for an OAuth 2.x/OpenID Connect client in the
    context of an authorization request.
    """

    authorize_url: str = pydantic.Field(
        default=...
    )

    client_id: str = pydantic.Field(
        default=...
    )

    created: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    params: AuthorizationRequestParameters = pydantic.Field(
        default=...
    )

    annotations: dict[str, str] = pydantic.Field(
        default_factory=dict
    )

    @property
    def pk(self) -> ClientAuthorizationStateKey:
        if self.params.state is None:
            raise NotImplementedError(
                "The authorization request did not include the `state` "
                "parameter."
            )
        return ClientAuthorizationStateKey(self.params.state)

    @property
    def redirect_uri(self) -> str | None:
        return self.params.redirect_uri

    def annotate(self, key: str, value: str) -> None:
        self.annotations[key] = value

    def annotation(
        self,
        key: str,
        decoder: Callable[[Any], T] = lambda x: x
    ) -> T | None:
        value = self.annotations.get(key)
        if value is None:
            return None

        return decoder(value)

    def get_authorize_url(self) -> str:
        return self.authorize_url

    async def verify(self, params: RedirectionParameters):
        if str(params.state) != str(self.params.state):
            raise TrustIssues(
                "The state returned by the authorization server "
                "does not match the expected state."
            )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from oauthx.server.handlers import BaseLoginRequestHandler
from oauthx.server.request import Request


class LoginRequestHandler(BaseLoginRequestHandler):
    __module__: str = 'oauthx.server.ref'

    async def handle(self, request: Request) -> fastapi.Response:
        if request.subject is None:
            return fastapi.Response(
                status_code=401,
                headers={'WWW-Authenticate': 'Basic'}
            )
        
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Annotated
from typing import TypeAlias

import fastapi
import google.auth.credentials
from aiopki.ext.google.client import get_credential
from google.cloud import datastore


__all__: list[str] = ['DatastoreClient']




async def get() -> datastore.Client:
    return datastore.Client(
        credentials=await get_credential(),
        project=os.environ['GOOGLE_DATASTORE_PROJECT'],
        namespace=os.environ['GOOGLE_DATASTORE_NAMESPACE']
    )


DatastoreClient: TypeAlias = Annotated[datastore.Client, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
import logging
from typing import Any
from typing import Callable
from typing import Coroutine

import fastapi
import fastapi.routing
import starlette.responses

from oauthx.lib.exceptions import InvalidRequest
from oauthx.lib.exceptions import InvalidToken
from oauthx.lib.exceptions import InsufficientScope


class ResourceRoute(fastapi.routing.APIRoute):
    logger: logging.Logger = logging.getLogger('uvicorn')

    def get_route_handler(self) -> Callable[[fastapi.Request], Coroutine[Any, Any, starlette.responses.Response]]:
        handler = super().get_route_handler()

        async def f(request: fastapi.Request) -> starlette.responses.Response:
            request = self.setup_request(request)
            try:
                return await handler(request)
            except (InvalidRequest, InvalidToken, InsufficientScope) as e:
                return fastapi.responses.PlainTextResponse(
                    status_code=e.status_code,
                    media_type='application/json;indent=2',
                    headers={
                        'WWW-Authenticate': str.join(',', [
                            f'error="{e.error}"',
                            f'error_description="{e.error_description}"'
                        ])
                    },
                    content=json.dumps({
                        'error': e.error,
                        'error_description': e.error_description
                    }, indent=2)
                )

        return f

    def setup_request(self, request: Any) -> Any:
        return request


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import dataclasses

from oauthx.server.models import SubjectKey
from oauthx.server.protocols import IRequestSubject
from oauthx.server.params import RequestSession


@dataclasses.dataclass
class RequestSubject:
    sub: str
    
    @property
    def pk(self) -> SubjectKey:
        return SubjectKey(self.sub)

    @classmethod
    async def resolve(cls, session: RequestSession) -> IRequestSubject | None:
        if session.sub is None:
            return None
        return cls(sub=session.sub)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import aiopki
import fastapi
import pydantic

from oauthx.lib.types import GrantType
from oauthx.lib.types import RedirectURI
from .confidentialclient import ConfidentialClient
from .publicclient import PublicClient


RESPONSE_TOKEN_TYPES: dict[str, str] = {
    'code'      : 'urn:ietf:params:oauth:token-type:access_token',
    'token'     : 'urn:ietf:params:oauth:token-type:access_token',
    'id_token'  : 'urn:ietf:params:oauth:token-type:id_token'
}


class BaseResponseMode(pydantic.BaseModel):
    client: ConfidentialClient | PublicClient
    redirect_uri: RedirectURI | None = None
    state: str | None = None
    iss: str

    @property
    def grant_type(self) -> GrantType | None:
        raise NotImplementedError

    def get_params(self) -> dict[str, str]:
        return self.model_dump(include={'state'}, exclude_none=True)

    def get_redirect_uri(self) -> RedirectURI:
        return self.redirect_uri or self.client.get_default_redirect_uri()

    def get_responses(self) -> set[str]:
        raise NotImplementedError

    def get_token_types(self) -> set[str]:
        return set(filter(bool, [RESPONSE_TOKEN_TYPES.get(x, '') for x in self.get_responses()]))

    async def deny(self) -> str:
        raise NotImplementedError

    async def error(
        self,
        *,
        error: str,
        error_description: str | None = None,
        error_uri: str | None = None
    ) -> str:
        raise NotImplementedError

    async def redirect(self, signer: aiopki.CryptoKeyType, **kwargs: Any) -> fastapi.responses.RedirectResponse:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .applicationcache import ApplicationCache
from .applicationstorage import ApplicationStorage
from .authorizationresponse import AuthorizationResponse
from .clientauthorizationstate import ClientAuthorizationState
from .currentsettings import CurrentSettings
from .currentsettings import CurrentSettingsDependency
from .defaultcache import DefaultCache
from .httpclientdependency import HTTPClient
from .httpclientdependency import HTTPClientDependency
from .logger import Logger
from .storage import Storage
from .templateservice import TemplateService
from .templateservice import TemplateServiceDependency


__all__: list[str] = [
    'ApplicationCache',
    'ApplicationStorage',
    'AuthorizationResponse',
    'ClientAuthorizationState',
    'CurrentSettings',
    'CurrentSettingsDependency',
    'DefaultCache',
    'HTTPClient',
    'HTTPClientDependency',
    'Logger',
    'Storage',
    'TemplateService',
    'TemplateServiceDependency',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Literal

import pydantic
from aiopki.ext.jose import CompactSerialized

from oauthx.lib.types import AccessToken
from oauthx.lib.types import ScopeType


class ObtainedGrant(pydantic.BaseModel):
    obtained: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    token_type: Literal['Bearer', 'bearer']
    expires_in: int
    access_token: AccessToken
    refresh_token: str | None = None
    id_token: CompactSerialized | None = None
    scope: ScopeType | None = None

    def is_expired(self) -> bool:
        """Return a boolean indicating if the grant is expired."""
        now = datetime.datetime.now(datetime.timezone.utc)
        return abs(int((now - self.obtained).total_seconds())) > self.expires_in


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from oauthx.lib.types import ResponseType
from oauthx.server.types import InvalidResponseType
from .confidentialclient import ConfidentialClient
from .responsemode import ResponseMode
from .publicclient import PublicClient


class AuthorizationContext(pydantic.BaseModel):
    client: ConfidentialClient | PublicClient
    response_mode: ResponseMode

    @property
    def response_type(self) -> ResponseType:
        return self.response_mode.response_type

    def can_redirect(self) -> bool:
        return self.response_mode.can_redirect()

    def model_post_init(self, __context: Any) -> None:
        if not self.client.allows_response_type(self.response_type):
            raise InvalidResponseType

    async def get_template_context(self) -> dict[str, Any]:
        return {
            'client_name': self.client.client_name,
            'deny_url': await self.deny()
        }
    async def deny(self) -> str:
        return await self.response_mode.error(error='access_denied')

    async def error(self, **params: Any) -> str:
        return await self.response_mode.error(**params)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from oauthx.lib.types import RedirectURI
from .iresourceowner import IResourceOwner


class IGrantedAuthorizationCode(Protocol):
    __module__: str = 'oauthx.types'
    value: str

    @property
    def owner(self) -> IResourceOwner.Key: ...

    def allows_redirect(self, redirect_uri: RedirectURI | None) -> bool: ...
    def consume(self) -> None: ...
    def is_authorized(self, client_id: str) -> bool: ...
    def is_consumed(self) -> bool: ...
    def has_consent(self, owner: IResourceOwner) -> bool: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import httpx
import pytest

from oauthx.models import Client


@pytest.fixture(scope='function')
def grant_types() -> list[str]:
    return ['client_credentials']


@pytest.mark.asyncio
async def test_client_credentials(
    agent: httpx.AsyncClient,
    http: httpx.AsyncClient,
    client: Client
):
    await client.client_credentials({'email'}, http)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any

import fastapi

from oauthx.lib.exceptions import InvalidScope
from oauthx.lib.exceptions import InvalidRequest
from oauthx.lib.models import AuthorizationCodeGrant
from oauthx.lib.models import ClientCredentialsGrant
from oauthx.lib.models import JWTBearerGrant
from oauthx.lib.models import Grant
from oauthx.lib.models import ObtainedGrant
from oauthx.server.params import ContentEncryptionKey
from oauthx.server.params import ObjectFactory
from oauthx.server.params import TokenIssuer
from oauthx.server.params import TokenSigner
from oauthx.server.request import Request
from oauthx.server.types import InvalidClient
from oauthx.server.types import UnauthorizedClient
from ..baserequesthandler import BaseRequestHandler
from .params import Assertion
from .params import Client
from .params import Grant
from .params import Authorization
from .params import ResourceOwner


class TokenEndpointHandler(BaseRequestHandler):
    __module__: str = 'oauthx.server.handlers'

    def setup(
        self,
        *,
        assertion: Assertion,
        cek: ContentEncryptionKey,
        client: Client,
        authorization: Authorization,
        factory: ObjectFactory,
        grant: Grant,
        issuer: TokenIssuer,
        signer: TokenSigner,
        owner: ResourceOwner,
        **_: Any
    ) -> None:
        self.assertion = assertion
        self.authorization = authorization
        self.cek = cek
        self.client = client
        self.factory = factory
        self.grant = grant
        self.issuer = issuer
        self.signer = signer
        self.owner = owner
        
        if grant.must_identify() and client is None:
            raise InvalidClient
        if client is not None and not client.can_grant(self.grant.grant_type):
            raise UnauthorizedClient
        if client is not None and not client.allows_scope(grant.scope):
            raise InvalidScope(
                "The client is not configured to allow the requested "
                "scope."
            )

    async def begin(self, request: Request) -> None:
        await self.grant.decrypt(NotImplemented)

    async def handle(self, request: Request) -> fastapi.Response:
        issued = await self.issue(self.grant.root)
        await self.on_grant_obtained(self.grant.root, issued)
        response = fastapi.responses.PlainTextResponse(
            media_type='application/json; indent=2',
            content=issued.model_dump_json(indent=2)
        )
        return response

    async def on_grant_obtained(self, grant: Any, obtained: ObtainedGrant):
        pass

    async def authorization_code(self, grant: AuthorizationCodeGrant) -> ObtainedGrant:
        assert self.authorization is not None
        assert self.client is not None
        assert self.owner is not None
        if not self.authorization.allows_scope(grant.scope):
            raise InvalidScope(
                "The requested scope exceeds the scope granted by "
                "the resource owner."
            )

        subject = await self.storage.get(self.owner.sub)
        at = await self.issuer.access_token(
            signer=self.signer,
            client_id=str(self.client.id),
            sub=str(self.owner.pk.sub),
            scope=self.authorization.scope
        )
        
        id_token = None
        if self.authorization.grants_id_token():
            await subject.decrypt_keys(self.cek)
            userinfo = await self.factory.userinfo(
                client=self.client,
                subject=subject,
                scope=grant.scope or self.authorization.scope,
                contributors=[self.authorization]
            )
            id_token = await self.issuer.id_token(
                signer=self.signer,
                userinfo=userinfo,
                access_token=at,
                code=grant.code
            )
        return ObtainedGrant(
            token_type='Bearer',
            expires_in=self.issuer.default_ttl,
            access_token=at,
            id_token=id_token
        )

    async def client_credentials(self, grant: ClientCredentialsGrant) -> ObtainedGrant:
        assert self.client is not None
        return ObtainedGrant(
            token_type='Bearer',
            expires_in=self.issuer.default_ttl,
            access_token=await self.issuer.access_token(
                signer=self.signer,
                client_id=str(self.client.id),
                sub=str(self.client.id),
                scope=grant.scope
            )
        )

    async def jwt(self, grant: JWTBearerGrant) -> ObtainedGrant:
        assert self.assertion is not None
        assert self.client is not None
        if self.assertion.is_self_issued():
            raise NotImplementedError

        owner = await self.storage.get(self.assertion.owner)
        if owner is None:
            raise InvalidRequest(
                "The resource owner identified by the assertion "
                "does not exist."
            )
        if not owner.grants_consent(grant.scope):
            raise InvalidScope(
                "The requested scope exceeds the scope granted by "
                "the resource owner."
            )

        return ObtainedGrant(
            token_type='Bearer',
            expires_in=self.issuer.default_ttl,
            access_token=await self.issuer.access_token(
                signer=self.signer,
                client_id=str(self.client.id),
                sub=self.assertion.sub,
                scope=grant.scope
            )
        )

    @functools.singledispatchmethod
    async def issue(self, grant: Any) -> ObtainedGrant:
        raise NotImplementedError

    @issue.register
    async def _(self, grant: AuthorizationCodeGrant) -> ObtainedGrant:
        return await self.authorization_code(grant)

    @issue.register
    async def _(self, grant: ClientCredentialsGrant) -> ObtainedGrant:
        return await self.client_credentials(grant)

    @issue.register
    async def _(self, grant: JWTBearerGrant) -> ObtainedGrant:
        return await self.jwt(grant)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

import fastapi

from oauthx.lib.types import GrantType


class IResponseMode(Protocol):
    __module__: str = 'oauthx.server.types'

    def get_token_types(self) -> set[str]: ...
    def grants(self) -> GrantType | None: ...

    async def deny(self) -> str: ...

    async def error(
        self,
        *,
        error: str,
        error_description: str | None = None,
        error_uri: str | None = None
    ) -> str:
        ...

    async def redirect(self, **kwargs: Any) -> fastapi.responses.RedirectResponse: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import inspect
import socket
from collections import OrderedDict
from typing import TypeVar
from typing import Any
from typing import Callable

import httpx


T = TypeVar('T')
EMPTY = inspect._empty # type: ignore


def http(func: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(func)
    async def f(self: Any, *args: Any, **kwargs: Any) -> Any:
        client = kwargs.pop('http', None)
        if client is not None:
            return await func(self, client, *args, **kwargs)
        async with httpx.AsyncClient() as client:
            return await func(self, client, *args, **kwargs)
    return f


def insert_signature(
    func: Callable[..., Any],
):
    def decorator_factory(wrapper: Callable[..., Any]) -> Callable[..., Any]:
        f = inspect.signature(func)
        w = inspect.signature(wrapper)
        wrapper.__signature__ = f.replace( # type: ignore
            parameters=[
                *w.parameters.values(),
                *f.parameters.values()
            ]
        )
        return wrapper

    return decorator_factory


def set_signature_defaults(
    callable: Callable[..., Any],
    defaults: dict[str, Any]
) -> Callable[..., Any]:
    sig = inspect.signature(callable)
    params = OrderedDict(sig.parameters.items())
    for name, default in defaults.items():
        if name not in params:
            continue
        params[name] = inspect.Parameter(
            kind=params[name].kind,
            name=params[name].name,
            default=default,
            annotation=params[name].annotation
        )

    async def f(*args: Any, **kwargs: Any) -> Any:
        return await callable(*args, **kwargs)
    f.__signature__ = sig.replace(parameters=list(params.values())) # type: ignore
    return f


def random_port() -> int:
    """Return an integer indicating a port on the local system that
    is available.
    """
    with socket.socket() as s:
        s.bind(('', 0))
        port = s.getsockname()[1]
    return port


class class_property:
    __module__: str = 'oauthx.utils'

    def __init__(self, func: Callable[..., Any]):
        self.func = func

    def __get__(self, instance: Any, cls: Any) -> Any:
        return self.func(cls)
    

def merge_signatures(signatures: list[inspect.Signature]) -> inspect.Signature:
    """Merge signatures to that FastAPI can inject the dependencies."""
    params: dict[str, inspect.Parameter] = OrderedDict()
    for sig in signatures:
        for param in sig.parameters.values():
            if param.name in {'self', 'cls'}:
                continue
            if param.name.startswith('_'):
                continue
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                continue
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                continue
            params[param.name] = param

    return signatures[0].replace(
        parameters=list(sorted(params.values(), key=lambda p: (p.kind, p.default != inspect._empty))) # type: ignore
    )


def merged_call(func: Callable[..., Any], kwargs: Any) -> Any:
    sig = inspect.signature(func)
    return func(**{k: v for k, v in kwargs.items() if k in sig.parameters})


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

import oauthx
from .types import AuthorizeFactory


@pytest.fixture(scope='function')
def scope() -> list[str]:
    return []


@pytest.mark.asyncio
async def test_client_must_allow_authorization_code(
    client: oauthx.Client,
    authorize_request: AuthorizeFactory
):
    await authorize_request(
        response_type='code',
        scope=['email'], prompt=['none'],
        expect_error='invalid_scope'
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Any
from typing import Union

import pydantic

from oauthx.lib.types import GrantType
from .authorizationcodegrant import AuthorizationCodeGrant
from .clientcredentialsgrant import ClientCredentialsGrant
from .jwtbearergrant import JWTBearerGrant
from .refreshtokengrant import RefreshTokenGrant
from .tokenexchangegrant import TokenExchangeGrant


__all__: list[str] = [
    'Grant'
]


_GrantType = Union[
    AuthorizationCodeGrant,
    ClientCredentialsGrant,
    JWTBearerGrant,
    RefreshTokenGrant,
    TokenExchangeGrant
]


class Grant(pydantic.RootModel[_GrantType]):
    root: _GrantType

    @property
    def grant_type(self) -> GrantType:
        return self.root.grant_type

    @property
    def scope(self) -> set[str]:
        if isinstance(self.root.scope, set):
            return self.root.scope
        return set(filter(bool, re.split(r'\s+', self.root.scope or '')))

    def has_credentials(self) -> bool:
        """Return a boolean indicating if client credentials were
        supplied with the grant.
        """
        return self.root.has_credentials()

    def must_authenticate(self) -> bool:
        """Return a boolean if the client must always authenticate when
        using this grant.
        """
        return self.root.must_authenticate()

    def must_identify(self) -> bool:
        """Return a boolean if the client must identify when
        using this grant.
        """
        return self.root.must_identify()

    async def decrypt(self, keychain: Any):
        pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import TypeVar

from canonical.protocols import ICache


T = TypeVar('T')


class MemoryCache(ICache):
    __module__: str = 'oauthx.lib'
    objects: dict[str, Any] = {}
    
    def __init__(self) -> None:
        self.objects = MemoryCache.objects

    async def get(
        self,
        key: str,
        decoder: Callable[[bytes], T] = bytes
    ) -> T | None:
        value = self.objects.get(key)
        if value is not None:
            return decoder(value)

    async def set(
        self,
        key: str,
        value: Any,
        encoder: Callable[..., bytes] = bytes,
        encrypt: bool = False
    ) -> None:
        self.objects[key] = encoder(value)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import dataclasses
from typing import Annotated
from typing import TypeAlias

import fastapi
from fastapi.security import HTTPBasic
from fastapi.security import HTTPBasicCredentials

from oauthx.lib.exceptions import InvalidRequest
from oauthx.server.models import ClientKey
from oauthx.server.models import ClientType
from .query import CLIENT_ID
from .query import CLIENT_SECRET


__all__: list[str] = ['ClientCredentials']


@dataclasses.dataclass
class ClientSecret:
    client_id: ClientKey
    client_secret: str | None

    async def authenticate(self, client: ClientType) -> bool:
        return all([
            client.id == str(self.client_id),
            await client.authenticate(self.client_secret)
        ])


async def get(
    credentials: HTTPBasicCredentials | None = fastapi.Depends(HTTPBasic(auto_error=False)),
    client_id: str | None = CLIENT_ID,
    client_secret: str | None = CLIENT_SECRET,
) -> ClientSecret | None:
    if credentials and (client_id or client_secret):
        raise InvalidRequest(
            "Multiple mechanisms for client authentication "
            "provided."
        )
    if not credentials and not client_id:
        return None
    if credentials:
        client_id = credentials.username
        client_secret = credentials.password
    assert client_id
    return ClientSecret(ClientKey(client_id), client_secret)


ClientCredentials: TypeAlias = Annotated[
    ClientSecret | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

import aiopki
import fastapi
import pydantic


P = TypeVar('P', bound=pydantic.BaseModel)
T = TypeVar('T')


class BaseSettings(pydantic.BaseModel):
    __module__: str = 'oauthx.lib.protocols'

    def inject(self) -> Any:
        def f(request: fastapi.Request):
            setattr(request, 'config', self)
        return fastapi.Depends(f)

    def get_cek(self) -> aiopki.CryptoKeyType:
        raise NotImplementedError

    def get_splash_image_url(self) -> str:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
import logging
from typing import Any
from typing import Callable
from typing import Iterable
from typing import TypeVar

import fastapi
import starlette.routing
from canonical.exceptions import ProgrammingError
from fastapi.types import DecoratedCallable

from oauthx.lib import utils
from oauthx.lib.exceptions import StopSnooping
from oauthx.lib.templates import TemplateService
from oauthx.lib.params import AuthorizationResponse
from oauthx.lib.params import ClientAuthorizationState
from oauthx.lib.params import Storage
from oauthx.lib.protocols import IClientStorage
from oauthx.lib.protocols import BaseSettings
from .clientapiroute import ClientAPIRoute
from .clientresolver import ClientResolver
from .params import ApplicationClient


T = TypeVar('T', bound=fastapi.FastAPI)


class ClientApplication(fastapi.FastAPI):
    __module__: str = 'oauthx.client'
    on_error_handler: Callable[..., Any] | None = None
    oauth: fastapi.APIRouter
    logger: logging.Logger = logging.getLogger('uvicorn')
    authorize_params: dict[str, Any]
    clients: ClientResolver
    scope: set[str]
    welcome_name: str
    welcome_path: str

    @classmethod
    def create_application(
        cls,
        *,
        settings: BaseSettings,
        client: ClientResolver,
        storage_class: type[IClientStorage],
        scope: Iterable[str] | None = None,
        login_path: str = '/login',
        login_name: str = 'oauth2.login',
        callback_path: str = '/oauth/v2/callback',
        callback_name: str = 'oauth2.callback',
        dependencies: list[Any] = [],
        auto_login: bool = False,
        redirect_root: bool = False,
        templates: TemplateService | None = None,
        template_dirs: list[str] | None = None,
        template_packages: list[str] | None = None,
        welcome_path: str = '/welcome',
        welcome_name: str = 'welcome',
        **kwargs: Any
    ):
        if templates is None:
            templates = TemplateService()
        app = cls(
            dependencies=[
                settings.inject(),
                templates.inject(
                    template_dirs=template_dirs,
                    packages=template_packages,
                ),
                storage_class.inject(),
                fastapi.Depends(client),
                *dependencies
            ],
            **kwargs
        )
        app.authorize_params = {}
        app.clients = client
        app.scope = set(scope or [])
        app.welcome_name = welcome_name
        app.welcome_path = welcome_path
        router = app.oauth = fastapi.APIRouter(route_class=ClientAPIRoute)
        router.add_api_route(
            methods=['GET'],
            path=callback_path,
            endpoint=app.callback,
            name=callback_name,
            dependencies=[
                client.as_provider(),
                *dependencies
            ],
            include_in_schema=False,
        )

        if auto_login:
            @app.get(login_path, include_in_schema=False)
            async def _(
                request: fastapi.Request,
            ) -> fastapi.Response:
                return fastapi.responses.RedirectResponse(
                    status_code=302,
                    url=request.url_for(login_name, preset='_')
                )

            @router.get(f'{login_path}/{{preset}}', name=login_name, include_in_schema=False)
            async def _(
                request: fastapi.Request,
                client: ApplicationClient,
                storage: Storage,
                preset: str = fastapi.Path(...),
            ) -> fastapi.Response:
                if preset not in app.authorize_params and preset != '_':
                    app.logger.warning("Authorization preset %s does not exist", preset[:32])
                    raise StopSnooping
                params = app.authorize_params.get(preset, {})
                state = await client.authorize(
                    redirect_uri=str(request.url_for(callback_name)),
                    **params
                )
                state.annotate('return-url', '/welcome')
                await storage.persist(state)
                return fastapi.responses.RedirectResponse(
                    status_code=302,
                    url=state.get_authorize_url()
                )

        if redirect_root:
            @router.get('/', include_in_schema=False)
            def _(request: fastapi.Request) -> fastapi.Response:
                try:
                    return fastapi.responses.RedirectResponse(
                        status_code=303,
                        url=request.url_for(login_name, preset='_')
                    )
                except starlette.routing.NoMatchFound:
                    raise ProgrammingError(
                        "No login endpoint is defined for this application. "
                        f"Use the {type(app).__name__}.on_login() decorator "
                        "to implement the login endpoint."
                    )

        app.include_router(router=router)
        return app

    async def callback(
        self,
        request: fastapi.Request,
        client: ApplicationClient,
        state: ClientAuthorizationState,
        response: AuthorizationResponse
    ) -> fastapi.Response:
        params = await client.on_redirected(state, response) # type: ignore
        name = 'oauth2.success'
        if params.is_error():
            name = 'oauth2.error'
        try:
            return fastapi.responses.RedirectResponse(
                status_code=302,
                url=request.url_for(name)\
                    .include_query_params(**request.query_params)
            )
        except starlette.routing.NoMatchFound:
            if params.is_error():
                return fastapi.responses.PlainTextResponse(
                    status_code=400,
                    content=f"{params.error}: {params.error_description}"
                )
            raise 

    def has_preset(self, name: str) -> bool:
        return name in self.authorize_params

    def login(self, preset: str, scope: Iterable[str] | None = None, **kwargs: Any) -> None:
        if preset == '_':
            raise ProgrammingError(
                "The _ preset is used for the default login endpoint "
                "if not preset is defined. It used the default settings "
                "specified for the client."
            )
        self.authorize_params[preset] = {
            **kwargs,
            'scope': set(scope or []) | self.scope
        }

    def on_authorized(
        self,
        preset: str,
        *,
        path_prefix: str = '/oauth/v2/success',
        name: str = 'oauth2.success'
    ) -> Callable[[DecoratedCallable], DecoratedCallable]:
        """Registers a function to handle a successful authorization."""
        if str.endswith(path_prefix, '/') or not str.startswith(path_prefix, '/'):
            raise ProgrammingError(
                "The path_prefix parameter must begin with a slash and not end "
                "with a slash."
            )
        if preset not in self.authorize_params:
            raise ProgrammingError(
                f"The preset '{preset}' has not been declared. "
                f"Call {type(self).__name__}.login() to declare "
                "a preset."
            )
        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            async def f(request: fastapi.Request, **kwargs: Any) -> fastapi.Response:
                response = utils.merged_call(func, {
                    **kwargs,
                    'request': request
                })
                if inspect.isawaitable(response):
                    response = await response
                params = {'channel': preset}
                if response is None:
                    response = fastapi.responses.RedirectResponse(
                        status_code=302,
                        url=request.url_for(self.welcome_name)\
                            .include_query_params(**params)
                    )
                return response

            f.__signature__ = utils.merge_signatures([ # type: ignore
                inspect.signature(f),
                inspect.signature(func)
            ])
            self.add_api_route(
                methods=['GET'],
                path=f'{path_prefix}/{preset}',
                endpoint=f,
                name=name,
                dependencies=[self.clients.as_provider()],
                include_in_schema=False
            )
            return func
        return decorator

    def welcome(self, func: DecoratedCallable) -> Callable[[DecoratedCallable], DecoratedCallable]:
        return self.get(self.welcome_path, name=self.welcome_name)(func)


create_application = ClientApplication.create_application


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any

import pydantic
import pytest

from oauthx.types import InvalidRequest
from oauthx.types import RedirectURI


VALID_VALUES: list[str] = [
    "https://google.com/callback",
    "http://127.0.0.1/callback",
    "http://127.0.0.1:8000/callback",
    "http://[::1]/callback",
    "http://[::1]:8000/callback",
    "https://google.com/",
    "http://127.0.0.1/",
    "http://127.0.0.1:8000/",
    "http://[::1]/",
    "http://[::1]:8000/",    
]


INVALID_VALUES: list[str | None] = [
    "invalid",
    None,
    'a' * 2049,
    'http://rp.example.com/callback',
    'http://localhost:8000/callback',
    'https://localhost:8000/callback',
    'https://localhost:8000/callback?foo=bar',
    'https://localhost:8000/callback#fragment',
    'http://1.1.1.1/callback',
    'http://::1',
    'https://::1',
    'http://0:0:0:0:0:0:0:1',
    'https://0:0:0:0:0:0:0:1',
    "urn:ietf:wg:oauth:2.0:oob",
    "urn:ietf:wg:oauth:2.0:oob:auto",
]


class M(pydantic.BaseModel):
    redirect_uri: RedirectURI

    class Config:
        json_encoders: dict[type, Any] = {
            RedirectURI: str
        }


@pytest.mark.parametrize("value", VALID_VALUES)
def test_valid_values(value: Any):
    RedirectURI(value)


@pytest.mark.parametrize("value", VALID_VALUES)
def test_valid_values_model(value: str):
    M.model_validate({'redirect_uri': value})


@pytest.mark.parametrize("value", INVALID_VALUES)
def test_invalid_values(value: Any):
    with pytest.raises(InvalidRequest):
        RedirectURI(value)


@pytest.mark.parametrize("value", INVALID_VALUES)
def test_invalid_values_model(value: str | None):
    if value is None:
        return
    with pytest.raises(pydantic.ValidationError):
        M.model_validate({'redirect_uri': value})


def test_cast_to_string():
    url = RedirectURI('https://rp.example.com')
    assert str(url) == 'https://rp.example.com'


def test_create_redirect_uri():
    url = RedirectURI('https://rp.example.com')
    p = urllib.parse.urlparse(url.redirect(foo='bar'))
    q = dict(urllib.parse.parse_qsl(p.query))
    assert 'foo' in q
    assert q.get('foo') == 'bar'


def test_serialize_model():
    obj = M.model_validate({
        'redirect_uri': 'https://rp.example.com/callback'
    })
    obj.model_dump_json()


@pytest.mark.parametrize("value", [
    "http://[::1]",
    "http://[::1]:8000",
    "http://[0:0:0:0:0:0:0:1]",
    "http://[0:0:0:0:0:0:0:1]:8000",
    "http://127.0.0.1",
    "http://127.0.0.1:8000",
])
def test_is_recognized_as_loopback(value: str):
    redirect_uri = RedirectURI(value)
    assert redirect_uri.is_loopback()


@pytest.mark.parametrize("x,y,allowed", [
    ("http://127.0.0.1", "http://127.0.0.1:8000", True),
    ("http://127.0.0.1:6000", "http://127.0.0.1:8000", True),
    ("http://[::1]:6000", "http://[::1]:8000", True),
    ("https://www.example.com/foo", "https://www.example.com/foo", True),
    ("https://www.example.com", "https://www.example.com:8000", False),
    ("https://www.example.com", "https://www.example.com/foo", False),
])
def test_allows_redirect(x: str, y: str, allowed: bool):
    assert RedirectURI(x).can_redirect(y) == allowed


def test_redirect_add_parameters():
    uri = RedirectURI('https://example.com')
    assert str.endswith(uri.redirect(foo=1), '?foo=1')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Any

import httpx
import pytest
from aiopki.ext import jose
from aiopki.types import ISigner

import oauthx
from .types import AuthorizeFactory


async def validate(
    agent: httpx.AsyncClient,
    http: httpx.AsyncClient,
    client: oauthx.Client,
    authorize_request: AuthorizeFactory,
    code: str,
    error: str
):
    _, params, state = await authorize_request(response_type='code')
    try:
        params.code = code # type: ignore
        await client.authorization_code(params, state, http=http)
        assert False
    except oauthx.Error as exc:
        assert exc.error == error


@pytest.mark.asyncio
@pytest.mark.parametrize("error,code", [
    ('invalid_request', os.urandom(16).hex()),
])
async def test_validate_authorization_code(
    agent: httpx.AsyncClient,
    http: httpx.AsyncClient,
    client: oauthx.Client,
    authorize_request: AuthorizeFactory,
    error: str,
    code: str
):
    await validate(agent, http, client, authorize_request, code, error)


@pytest.mark.asyncio
@pytest.mark.parametrize("params", [
    {'aut': 1, 'mac': '', 'sub': '1'}
])
async def test_validate_authorization_code_signature(
    agent: httpx.AsyncClient,
    http: httpx.AsyncClient,
    client: oauthx.Client,
    authorize_request: AuthorizeFactory,
    random_signer: ISigner,
    params: dict[str, Any]
):
    params.setdefault('client_id', str(client.client_id))
    jws = jose.jws(params)
    await jws.sign(random_signer.default_algorithm(), random_signer)
    await validate(agent, http, client, authorize_request, jws.encode(bytes.decode), 'invalid_grant')


@pytest.mark.asyncio
@pytest.mark.parametrize("params", [
    {},
])
async def test_validate_authorization_code_payload(
    agent: httpx.AsyncClient,
    http: httpx.AsyncClient,
    client: oauthx.Client,
    authorize_request: AuthorizeFactory,
    random_signer: ISigner,
    params: dict[str, Any]
):
    params.setdefault('client_id', str(client.client_id))
    jws = jose.jws(params)
    await jws.sign(random_signer.default_algorithm(), random_signer)
    await validate(agent, http, client, authorize_request, jws.encode(bytes.decode), 'invalid_grant')


@pytest.mark.asyncio
@pytest.mark.parametrize("params", [
    {'aut': 1, 'mac': '', 'sub': '1'}
])
async def test_validate_authorization_code_mac(
    agent: httpx.AsyncClient,
    http: httpx.AsyncClient,
    client: oauthx.Client,
    authorize_request: AuthorizeFactory,
    signer: ISigner,
    params: dict[str, Any]
):
    params.setdefault('client_id', str(client.client_id))
    jws = jose.jws(params)
    await jws.sign(signer.default_algorithm(), signer)
    await validate(agent, http, client, authorize_request, jws.encode(bytes.decode), 'invalid_grant')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os

import pytest

from .types import AuthorizeFactory


@pytest.mark.asyncio
@pytest.mark.parametrize("redirect_uri", [
    f'http://127.0.0.1:23390/{os.urandom(16).hex()}'
])
async def test_invalid_redirect_uri_does_not_redirect(
    redirect_uri: str,
    authorize_request: AuthorizeFactory
):
    await authorize_request(
        redirect_uri=redirect_uri,
        status_code=403
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
from aiopki.ext.jose import CompactSerialized

from oauthx.lib.exceptions import InvalidScope
from oauthx.lib.models import Grant as GrantModel
from oauthx.lib.types import GrantType
from oauthx.lib.types import RedirectURI
from .granttype import GrantType
from .resourceowner import ResourceOwner
from .query import ASSERTION
from .query import CODE
from .query import REDIRECT_URI
from .query import SCOPE


__all__: list[str] = [
    'Grant'
]


async def get(
    grant_type: GrantType,
    owner: ResourceOwner,
    assertion: CompactSerialized | None = ASSERTION,
    code: CompactSerialized = CODE,
    client_id: str | None = fastapi.Form(
        default=None,
        title="Client ID",
        description=(
            "Required if the client is not authenticating with "
            "the authorization server"
        )
    ),
    client_secret: str | None = fastapi.Form(
        default=None,
        title="Client secret",
        description=(
            "This parameter is **required** if the client is authenticating using "
            "the `client_secret_post` method, otherwise is **must** be "
            "omitted."
        )
    ),
    redirect_uri: RedirectURI | None = REDIRECT_URI,
    scope: str | None = SCOPE
) -> GrantModel:
    grant = GrantModel.model_validate({
        'grant_type': grant_type,
        'assertion': assertion,
        'client_id': client_id,
        'client_secret': client_secret,
        'code': code.encode() if code else None,
        'redirect_uri': redirect_uri,
        'scope': scope
    })
    await grant.decrypt(NotImplemented)
    if owner is not None and not owner.grants_consent(grant.scope):
        raise InvalidScope(
            "The resource owner has revoked consent "
            "for the requested scope."
        )

    return grant


Grant: TypeAlias = Annotated[
    GrantModel,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from .clientauthorizationstate import ClientAuthorizationState
from .error import Error
from .jarmtoken import JARMToken
from .jarmauthorizationresponse import JARMAuthorizationResponse
from .provider import Provider
from .redirectionparameters import RedirectionParameters


class AuthorizationResponse(pydantic.RootModel[Error | JARMToken | JARMAuthorizationResponse | RedirectionParameters]):

    @property
    def code(self):
        assert isinstance(self.root, RedirectionParameters)
        return self.root.code

    @property
    def iss(self) -> str | None:
        if not isinstance(self.root, RedirectionParameters):
            return None
        return self.root.iss

    def can_retry(self) -> bool:
        return isinstance(self.root, Error) and self.root.error in {
            'access_denied'
        }

    def is_error(self) -> bool:
        return isinstance(self.root, Error)

    def is_jarm(self) -> bool:
        return isinstance(self.root, JARMToken)

    def raise_for_status(self):
        return self.root.raise_for_status()

    async def decode(self, issuer: str, client_id: str, verifier: Any, decrypter: Any):
        if isinstance(self.root, JARMToken):
            self.root = await self.root.decode(issuer, client_id, verifier, decrypter)

    async def verify(self, provider: Provider, state: ClientAuthorizationState):
        self.root.raise_for_status()
        assert isinstance(self.root, RedirectionParameters)
        await state.verify(self.root)
        await provider.verify(self.root)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.server import models
from oauthx.server.params import Storage
from oauthx.server.request import Request
from .client import Client
from .redirecturi import RedirectURI
from .responsemode import ResponseMode


__all__: list[str] = [
    'AuthorizationContext'
]


async def get(
    request: Request,
    storage: Storage,
    client: Client,
    redirect_uri: RedirectURI,
    response_mode: ResponseMode
) -> models.AuthorizationContext:
    ctx = models.AuthorizationContext.model_validate({
        'client': client,
        'redirect_uri': redirect_uri,
        'response_mode': response_mode,
    })
    request.set_context(ctx)
    return ctx


AuthorizationContext: TypeAlias = Annotated[
    models.AuthorizationContext,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Protocol
from typing import MutableMapping

import httpx
from aiopki.ext.jose import JOSEObject
from canonical import ResourceIdentifier
from fastapi.security import HTTPBasicCredentials

from oauthx.lib.types import GrantType
from oauthx.lib.types import RedirectURI
from oauthx.lib.types import ResponseType
from .irequestsubject import IRequestSubject
from .iresourceowner import IResourceOwner


class IClient(Protocol):
    __module__: str = 'oauthx.types'

    @property
    def id(self) -> str:
        ...

    def allows_response_type(self, response_type: ResponseType) -> bool:
        """Return a boolean indicating if the client allows the given
        response type.
        """
        ...

    def allows_scope(self, scope: set[str]) -> bool:
        """Return a boolean indicating if the client allows the
        given scope.
        """
        ...

    def can_redirect(self, redirect_uri: RedirectURI | None) -> bool:
        """Return a boolean indicating if the client allows
        redirection to the given `redirect_uri`.
        """
        ...

    def can_grant(self, grant_type: GrantType | None) -> bool:
        """Return a boolean indicating if the client allows the given
        `grant_type`.
        """
        ...

    def contribute_to_userinfo(self, userinfo: MutableMapping[str, Any]) -> None:
        ...

    def get_default_redirect_uri(self) -> RedirectURI:
        """Return a string containing the default redirect URI for the
        client. Raise :exc:`NotImplementedError` if the client can not
        redirect.
        """
        ...

    def get_display_name(self) -> str:
        """Return a string containing the display name of the client."""
        ...

    def get_logo_url(self) -> str:
        """Return a string containing the logo URL of the client."""
        ...

    def get_sector_identifier(self) -> str: ...

    def is_confidential(self) -> bool:
        """Return a boolean indicating if the client is confidential."""
        ...

    def must_push(self) -> bool:
        """Return a boolean indicating if the client is required to
        push authorization requests.
        """
        ...

    def resource_owner(self, subject: IRequestSubject) -> IResourceOwner.Key:
        """Create a resource owner identifier."""
        ...

    def requires_state(self) -> bool:
        """Return a boolean indicating if the client enforces the use
        of the ``state`` parameter in authorization requests.
        """
        ...

    def update_key(self, update: Callable[[bytes], None]):
        """Update the clients' key that is used to secure its
        security credentials, such as authorization codes and
        refresh tokens. Updating the key effectively revokes
        all security credentials for the client.
        """
        ...

    async def authenticate(self, credential: HTTPBasicCredentials | JOSEObject | str | None) -> None:
        """Authenticate the client using the given credential.
        """
        ...

    async def userinfo(self, access_token: str, http: httpx.AsyncClient | None = None) -> None:
        """Query the providers' UserInfo endpoint to obtain information
        about the resource owner.
        """
        ...

    class Key(ResourceIdentifier[str, 'IClient']):
        client_id: str
        openapi_example: str = 'client-123'
        openapi_title: str = 'Client ID'

        def __init__(self, client_id: str):
            self.client_id = client_id

        def cast(self) -> str:
            return str(self.client_id)

        def __str__(self) -> str:
            return self.client_id
        
        def __eq__(self, key: object) -> bool:
            return isinstance(key, type(self)) and key.client_id == self.client_id


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import httpx
import pytest

import oauthx


@pytest.fixture
def grant_types() -> list[str]:
    return []


@pytest.fixture
def redirect_uris() -> list[str]:
    return []


@pytest.mark.asyncio
async def test_missing_redirect_uri_does_not_redirect(
    agent: httpx.AsyncClient,
    client: oauthx.Client,
):
    response = await agent.get(
        url='https://127.0.0.1/oauth/v2/authorize',
        params={
            'client_id': client.client_id,
            'response_type': 'code',
        }
    )
    assert response.status_code == 400, response.headers


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
import re
import pathlib

import aiofiles
import yaml

from .models import Client


class LocalClientStorage:
    __module__: str = 'oauthx.server'
    base_dir: pathlib.Path = pathlib.Path('etc/clients.d')
    logger: logging.Logger = logging.getLogger('uvicorn')
    pattern: re.Pattern[str] = re.compile('^[a-zA-Z0-9]([a-zA-Z0-9.]+)[a-zA-Z0-9]$')

    async def get(self, client_id: str) -> Client | None:
        if not self.pattern.match(str(client_id) or ''):
            return None
        path = self.base_dir.joinpath(f'{client_id}.yaml')
        if not path.exists():
            self.logger.debug(
                "Client does not exist in local configuration (client: %s)",
                client_id
            )
            return None
        try:
            async with aiofiles.open(path) as f:
                params = yaml.safe_load(await f.read())
            return Client.model_validate({**params, 'client_id': client_id})
        except Exception as e:
            self.logger.critical(
                "Caught fatal %s while retrieving client from "
                "local configuration  (client: %s)",
                type(e).__name__, client_id
            )
            return None
        


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .authorizationcontext import AuthorizationContext
from .authorizationrequest import AuthorizationRequest
from .client import Client
from .redirecturi import RedirectURI
from .responsemode import ResponseMode
from .responsetype import ResponseType
from .resourceowner import ResourceOwner
from .scope import Scope
from .state import State


__all__: list[str] = [
    'AuthorizationContext',
    'AuthorizationRequest',
    'Client',
    'RedirectURI',
    'ResponseMode',
    'ResponseType',
    'ResourceOwner',
    'Scope',
    'State',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from oauthx.lib.types import TokenType
from .basegrant import BaseGrant


class TokenExchangeGrant(BaseGrant):
    grant_type: Literal['urn:ietf:params:oauth:grant-type:token-exchange'] = pydantic.Field(
        default=...,
        description=(
            "The value `urn:ietf:params:oauth:grant-type:token-exchange` "
            "indicates that a token exchange is being performed."
        )
    )

    resource: str | None = pydantic.Field(
        default=None,
        description=(
            "A URI that indicates the target service or resource where "
            "the client intends to use the requested security token. "
            "This enables the authorization server to apply policy as "
            "appropriate for the target, such as determining the type "
            "and content of the token to be issued or if and how the "
            "token is to be encrypted. In many cases, a client will "
            "not have knowledge of the logical organization of the "
            "systems with which it interacts and will only know a URI "
            "of the service where it intends to use the token. The "
            "resource parameter allows the client to indicate to the "
            "authorization server where it intends to use the issued "
            "token by providing the location, typically as an https "
            "URL, in the token exchange request in the same form that "
            "will be used to access that resource. The authorization "
            "server will typically have the capability to map from a "
            "resource URI value to an appropriate policy. The value "
            "of the resource parameter MUST be an absolute URI that "
            "**may** include a query component and **must not** "
            "include a fragment component. Multiple resource "
            "parameters may be used to indicate that the issued "
            "token is intended to be used at the multiple resources "
            "listed."
        )
    )

    audience: str | None = pydantic.Field(
        default=None,
        description=(
            "The logical name of the target service where the client "
            "intends to use the requested security token. This serves "
            "a purpose similar to the resource parameter but with the "
            "client providing a logical name for the target service. "
            "Interpretation of the name requires that the value be "
            "something that both the client and the authorization "
            "server understand. An OAuth client identifier, a SAML "
            "entity identifier, and an OpenID Connect Issuer Identifier "
            "are examples of things that might be used as audience "
            "parameter values. However, audience values used with a "
            "given authorization server must be unique within that "
            "server to ensure that they are properly interpreted as "
            "the intended type of value. Multiple audience parameters "
            "may be used to indicate that the issued token is intended "
            "to be used at the multiple audiences listed. The audience "
            "and resource parameters may be used together to indicate "
            "multiple target services with a mix of logical names and "
            "resource URIs."
        )
    )

    scope: str | None = pydantic.Field(
        default=None,
        description=(
            "A list of space-delimited, case-sensitive strings, that "
            "allow the client to specify the desired scope of the "
            "requested security token in the context of the service "
            "or resource where the token will be used. The values and "
            "associated semantics of scope are service specific and "
            "expected to be described in the relevant service documentation."
        )
    )

    requested_token_type: TokenType | None = pydantic.Field(
        default=None,
        title="Requested token type",
        description=(
            "An identifier for the type of the requested security "
            "token. If the requested type is unspecified, the issued "
            "token type is at the discretion of the authorization "
            "server and may be dictated by knowledge of the requirements "
            "of the service or resource indicated by the resource or "
            "audience parameter."
        )
    )

    subject_token: str = pydantic.Field(
        default=...,
        title="Subject token",
        description=(
            "A security token that represents the identity of the "
            "party on behalf of whom the request is being made. "
            "Typically, the subject of this token will be the "
            "subject of the security token issued in response to "
            "the request."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.params import Storage
from oauthx.server.config import Config
from oauthx.server.models import ObjectFactory as DefaultImplementation
from .currentconfig import CurrentConfig
from .issueridentifier import IssuerIdentifier
from .masker import Masker


__all__: list[str] = ['ObjectFactory']


async def get(
    issuer: IssuerIdentifier,
    masker: Masker,
    storage: Storage,
    config: Config = CurrentConfig
) -> DefaultImplementation:
    return DefaultImplementation(
        issuer=issuer,
        masker=masker, # type: ignore
        storage=storage,
        scopes=[x.scope for x in config.scopes],
    )


ObjectFactory: TypeAlias = Annotated[DefaultImplementation, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any
from typing import Literal
from typing import MutableMapping
from typing import TypeAlias
from typing import Union
from typing_extensions import Literal

import pydantic
from aiopki.types import EncryptionResult
from aiopki.types import ISigner

from oauthx.server.protocols import ISubject
from .birthdate import BirthdateClaim
from .email import EmailClaim
from .gender import GenderClaim
from .name import NameClaim
from .phonenumber import PhonenumberClaim
from .stringclaim import StringClaim


__all__: list[str] = ['Claim']


ClaimType: TypeAlias = Union[
    BirthdateClaim,
    EmailClaim,
    GenderClaim,
    NameClaim,
    PhonenumberClaim,
    StringClaim,
]


class Claim(pydantic.RootModel[ClaimType]):

    @classmethod
    def new(
        cls,
        *,
        receipt_id: int,
        id: int,
        kind: Literal['email', 'gender', 'name'],
        provider: str,
        issuer: str,
        sub: int,
        value: Any,
        now: datetime.datetime | None = None,
        ial: int = 0
    ):
        """Create a new :class:`Claim` instance using an
        Open ID Connect ID Token.
        """
        return cls.model_validate({
            'ial': ial,
            'id': id,
            'issuer': issuer,
            'kind': kind,
            'obtained': now or datetime.datetime.now(datetime.timezone.utc),
            'provider': provider,
            'receipt_id': receipt_id,
            'sub': sub,
            'value': value,
        })

    @property
    def pk(self) -> int:
        return self.root.id

    def contribute_to_userinfo(self, userinfo: MutableMapping[str, Any]) -> None:
        self.root.contribute_to_userinfo(userinfo)

    def is_encrypted(self) -> bool:
        """Return a boolean indicating if the claim is encrypted."""
        return isinstance(self.root.value, EncryptionResult)

    def is_masked(self) -> bool:
        """Return a boolean indicating if the claim is masked."""
        return self.root.is_masked()

    async def decrypt(self, subject: ISubject) -> None:
        """Decrypt the claim using the owners' private key."""
        return await self.root.decrypt(subject)

    async def encrypt(self, subject: ISubject) -> None:
        """Encrypt the claim using the owners' private key."""
        await self.root.encrypt(subject)

    async def mask(self, key: ISigner) -> None:
        """Use `key` to create a masked and set the :class:`BaseClaim.masked`
        attribute.
        """
        await self.root.mask(key)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic import GetCoreSchemaHandler
from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue


__all__: list[str] = [
    'ClientSecret'
]


class ClientSecret(str):
    __module__: str = 'oauthx.types'

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _: Any,
        handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        
        def is_not_uri(client_secret: str):
            # TODO: this assumes that a client secret never contains a colon.
            p = urllib.parse.urlparse(client_secret)
            if p.netloc or p.scheme:
                raise ValueError("A client secret can not be an URI.")
            return client_secret

        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(max_length=128),
            python_schema=core_schema.chain_schema([
                core_schema.is_instance_schema(str),
                core_schema.no_info_plain_validator_function(is_not_uri),
                core_schema.no_info_plain_validator_function(cls)
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema(max_length=128))
    
    def __await__(self):
        async def f(): return self
        return f().__await__()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical import ResourceIdentifier

from oauthx.lib.protocols import IClient


class ClientKey(ResourceIdentifier[int | str, IClient]):
    __module__: str = 'oauthx.types'
    
    def __init__(self, client_id: str):
        self.client_id = client_id

    def __str__(self) -> str:
        return self.client_id

    def __hash__(self) -> int:
        return hash(self.client_id)

    def __repr__(self) -> str:
        return f'<ClientKey: {str(self)}>'
    
    def __eq__(self, key: object) -> bool:
        return all([
            isinstance(key, type(self)),
            hash(self) == hash(key)
        ])


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any
from typing import Literal

import pydantic

from oauthx.server.protocols import ISubject
from oauthx.lib.types import MaskedPrincipal
from oauthx.lib.types import OIDCIssuerIdentifier


class BasePrincipal(pydantic.BaseModel):
    active: bool = pydantic.Field(
        default=True
    )

    allow_login: bool = pydantic.Field(
        default=True
    )

    blocked: bool = pydantic.Field(
        default=False
    )

    issuer: Literal['self'] | OIDCIssuerIdentifier = pydantic.Field(
        default=...
    )

    kind: Any = pydantic.Field(
        default=...
    )

    masked: MaskedPrincipal = pydantic.Field(
        default=...
    )

    registered: datetime.datetime = pydantic.Field(
        default=...
    )
    
    owner: int | str = pydantic.Field(
        default=...
    )
    
    verified: bool = pydantic.Field(
        default=False
    )

    value: Any

    def to_bytes(self) -> bytes:
        raise NotImplementedError

    async def encrypt(self, subject: ISubject) -> None:
        self.value = await subject.encrypt(bytes(self))

    def __bytes__(self) -> bytes:
        return self.to_bytes()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi
from aiopki.ext.jose import CompactSerialized

from oauthx.lib.types import GrantType
from oauthx.lib.types import RedirectURI
from oauthx.lib.types import ScopeType


__all__: list[str] = [
    'CLIENT_ID',
    'CLIENT_SECRET',
    'CODE',
    'GRANT_TYPE',
    'REDIRECT_URI',
]


ASSERTION: CompactSerialized | None = fastapi.Form(
    default=None,
    description=(
        "A compact-encoded JSON Web Token (JWT) that is signed, "
        "containing the claims specified in RFC 7523."
    )
)

CLIENT_ID: Any = fastapi.Form(
    default=None,
    title="Client ID",
    description=(
        "Required if the client is not authenticating with "
        "the authorization server"
    )
)

CLIENT_SECRET: str | None = fastapi.Form(
    default=None,
    title="Client secret",
    description=(
        "This parameter is **required** if the client is authenticating using "
        "the `client_secret_post` method, otherwise is **must** be "
        "omitted."
    )
)

CODE: CompactSerialized | Any = fastapi.Form(
    default=None,
    description=(
        "The authorization code received from the authorization "
        "server when using the `authorization_code` grant."
    )
)

GRANT_TYPE: GrantType = fastapi.Form(
    default=...,
    description="Specifies the grant to obtain."
)

REDIRECT_URI: RedirectURI | None = fastapi.Form(
    default=None,
    title="Redirect URI",
    description=(
        "Required if the `redirect_uri` parameter was included "
        "in the authorization request, and their values must "
        "be identical."
    )
)

SCOPE: str | None = fastapi.Form(
    default_factory=ScopeType,
    title="Scope",
    description=(
        "The requested access token scope. The resource owner "
        "**must** allow the requested scope."
    )
)



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
from aiopki.ext.jose import CompactSerialized
from aiopki.ext.jose import JWT

from oauthx.lib.exceptions import InvalidGrant
from oauthx.lib.types import GrantType
from oauthx.server import models
from .query import ASSERTION
from .query import GRANT_TYPE


__all__: list[str] = [
    'Assertion'
]


MAX_AGE: int = 600


async def get(
    request: fastapi.Request,
    grant_type: GrantType = GRANT_TYPE,
    assertion: CompactSerialized | None = ASSERTION,
) -> models.Assertion | None:
    if assertion is None:
        return
    aud = str(request.url_for('oauth2.token'))
    jwt = assertion.payload(JWT.model_validate)
    if not jwt.validate_aud(aud):
        raise InvalidGrant(
            "The audience specified by the assertion is invalid. "
            "An assertion must point to the authorization servers' "
            f"token endpoint ({aud})."
        )

    if not all([jwt.validate_exp(), jwt.validate_nbf()]):
        raise InvalidGrant(
            "The provided assertion can not be used at the current "
            "date and time."
        )

    if not jwt.validate_iat(MAX_AGE):
        raise InvalidGrant("The provided assertion is too old.")

    
    return models.Assertion.model_validate({
        **jwt.model_dump(mode='json'),
        'grant_type': grant_type,
        'jws': assertion
    })


Assertion: TypeAlias = Annotated[
    models.Assertion | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
from typing import Any
from typing import Callable

import pydantic
from canonical import EmailAddress


class UserInfoResponse(pydantic.BaseModel):
    claims: dict[str, Any] = pydantic.Field(default={}, alias='__claims__')
    email: EmailAddress | None = None
    email_verified: bool | None = None
    family_name: str | None = None
    given_name: str | None = None
    middle_name: str | None = None
    name: str | None = None
    sub: str

    @pydantic.model_validator(mode='before')
    def preprocess(cls, values: bytes | str | dict[str, Any]) -> dict[str, Any]:
        if isinstance(values, (bytes, str)):
            values = json.loads(values)
        assert isinstance(values, dict)
        return {
            **values,
            '__claims__': {
                k: v
                for k, v in values.items()
                if k not in cls.model_fields
            }
        }

    @pydantic.model_serializer(mode='wrap', when_used='always')
    def serialize_model(
        self,
        info: Callable[['UserInfoResponse'], dict[str, Any]]
    ):
        values = info(self)
        if 'claims' in values:
            values = {**values.pop('claims'), **values}
        return values

    def get(self, claim: str) -> Any:
        return self.claims.get(claim)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from oauthx.lib.types import ScopeType
from .basegrant import BaseGrant


class ClientCredentialsGrant(BaseGrant):
    grant_type: Literal['client_credentials'] = pydantic.Field(
        default=...,
        description="This value **must** be `client_credentials`."
    )

    scope: ScopeType = pydantic.Field(
        default_factory=ScopeType,
        description=(
            "The space-delimited scope of the access request. The client "
            "**must** allow the given scope."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from aiopki.types import EncryptionResult

from .baseclaim import BaseClaim


class IntegerClaim(BaseClaim):
    kind: str
    value: int | EncryptionResult

    def get_encryption_input(self) -> bytes:
        assert isinstance(self.value, str)
        return str.encode(self.value, 'utf-8')

    @BaseClaim.requires_decryption
    def get_masking_input(self) -> bytes:
        assert isinstance(self.value, str)
        return str.encode(self.value, 'utf-8')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import Any
from typing import AsyncGenerator
from typing import TypeAlias

import fastapi
from aiopki.ext.jose import CompactSerialized

from oauthx.lib.exceptions import InvalidGrant
from oauthx.lib.exceptions import UnknownClient
from oauthx.lib.params import Logger
from oauthx.lib.types import GrantType
from oauthx.lib.types import RedirectURI
from oauthx.server.models import Authorization as AuthorizationState
from oauthx.server.models import AuthorizationCode
from oauthx.server.params import Storage
from oauthx.server.params import TokenSigner
from .client import Client
from .resourceowner import ResourceOwner
from .query import CODE
from .query import GRANT_TYPE
from .query import REDIRECT_URI


__all__: list[str] = [
    'Authorization'
]


async def get(
    signer: TokenSigner,
    storage: Storage,
    logger: Logger,
    client: Client,
    owner: ResourceOwner,
    grant_type: GrantType = GRANT_TYPE,
    code: CompactSerialized = CODE,
    redirect_uri: RedirectURI | None = REDIRECT_URI
) -> AsyncGenerator[Any, AuthorizationState | None]:
    if grant_type != 'authorization_code':
        yield None
        return
    if client is None:
        raise UnknownClient(
            "No client could be identified from the Authorization "
            "header, the request parameters, or client assertion."
        )
    if owner is None:
        raise InvalidGrant(
            "The resource owner identified by the grant parameters "
            "does not exist."
        )
    if not await code.verify(signer):
        raise InvalidGrant("The provided authorization code is not valid.")
    granted = code.payload(AuthorizationCode.model_validate_json)
    
    if not granted.is_authorized(str(client.id)): # type: ignore
        raise InvalidGrant("The client is not authorized to obtain this grant.")
    if granted.is_revoked(client, owner):
        raise InvalidGrant("This authorization code is revoked.")
    if not granted.allows_redirect(redirect_uri):
        raise InvalidGrant(
            "The redirect_uri parameter must match the redirect_uri that "
            "was included in the authorization request."
        )
    authorization = await storage.get(granted.authorization)
    if authorization is None or authorization.is_consumed():
        if authorization is not None:
            logger.critical(
                "Authorization code reused (authorization: %s)",
                str(authorization.pk)
            )
        logger.debug("No authorization exists")
        raise InvalidGrant(
            "The authorization code supplied in the request body can not "
            "be used."
        )
    yield authorization
    await authorization.consume(storage)


Authorization: TypeAlias = Annotated[
    AuthorizationState | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from oauthx.lib.params import ClientAuthorizationState

from .authorizationserver import AuthorizationServer
from .contentencryptionkey import ContentEncryptionKey
from .currentconfig import CurrentConfig
from .currentsubject import CurrentSubject
from .datamaskingkey import DataMaskingKey
from .httpclientdependency import HTTPClientDependency
from .issuer import Issuer
from .issueridentifier import IssuerIdentifier
from .masker import Masker
from .objectfactory import ObjectFactory
from .obtainedgrant import ObtainedGrant
from .oidctoken import OIDCToken
from .pluginrunner import PluginRunner
from .pendingauthorizationclient import PendingAuthorizationClient
from .pendingauthorizationrequest import PendingAuthorizationRequest
from .requestingclient import RequestingClient
from .requestsession import RequestSession
from .returnurl import ReturnURL
from .sessionsigner import SessionSigner
from .storage import Storage
from .subjectlogger import SubjectLogger
from .tokenissuer import TokenIssuer
from .tokensigner import TokenSigner
from .upstreamprovider import UpstreamProvider


__all__: list[str] = [
    'AuthorizationServer',
    'ClientAuthorizationState',
    'ContentEncryptionKey',
    'CurrentConfig',
    'CurrentSubject',
    'DataMaskingKey',
    'HTTPClientDependency',
    'Issuer',
    'IssuerIdentifier',
    'Masker',
    'ObjectFactory',
    'ObtainedGrant',
    'OIDCToken',
    'PluginRunner',
    'PendingAuthorizationClient',
    'PendingAuthorizationRequest',
    'RequestingClient',
    'RequestSession',
    'ReturnURL',
    'SessionSigner',
    'Storage',
    'SubjectLogger',
    'TokenIssuer',
    'TokenSigner',
    'UpstreamProvider'
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any


class Error(Exception):
    __module__: str = 'headless.ext.oidc.types'
    allow_redirect: bool = True
    error: str
    error_description: str | None = None
    error_uri: str | None = None
    media_type: str | None = None
    status_code: int = 400

    @classmethod
    def invalid_request(cls, description: str):
        return cls(error='invalid_request', error_description=description)

    def __init__(
        self,
        *,
        error: str,
        error_description: str | None,
        error_uri: str | None = None,
        allow_redirect: bool = False,
        state: str | None = None,
        **kwargs: Any
    ):
        self.allow_redirect = allow_redirect
        self.error = error
        self.error_description = error_description
        self.error_uri = error_uri
        self.state = state

    def can_redirect(self) -> bool:
        return self.allow_redirect

    def dict(self) -> dict[str, str]:
        dto = {'error': self.error}
        if self.error_description:
            dto['error_description'] = self.error_description
        if self.error_uri:
            dto['error_uri'] = self.error_uri
        return dto

    def fatal(self) -> bool:
        return False

    def get_context(self) -> Any:
        return {}

    def get_template_names(self) -> list[str]:
        return [f'oauthx/errors/error.{self.error}.html.j2']

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"Error('{self.error_description or 'None'}')"


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any
from typing import Callable
from typing import Literal
from typing import MutableMapping
from typing import TypeVar

import pydantic
from aiopki.ext.jose import OIDCToken
from aiopki.lib import JSONWebKey
from aiopki.types import IDecrypter
from aiopki.types import IEncrypter
from aiopki.types import EncryptionResult
from aiopki.types import Plaintext
from aiopki.utils import b64encode

from .subjectkey import SubjectKey


R = TypeVar('R')


class Subject(pydantic.BaseModel):
    active: bool = pydantic.Field(
        default=True
    )    

    dek: EncryptionResult | JSONWebKey = pydantic.Field(
        default=...
    )

    dmk: EncryptionResult | JSONWebKey = pydantic.Field(
        default=...
    )

    kind: Literal['User'] = pydantic.Field(
        default=...
    )

    locale: str = pydantic.Field(
        default='nl-NL'
    )

    sub: int = pydantic.Field(
        default=...
    )

    use: Literal['personal', 'institutional'] = pydantic.Field(
        default=...
    )

    zoneinfo: str = pydantic.Field(
        default='Europe/Amsterdam'
    )

    @staticmethod
    def requires_decryption(func: Callable[..., R]) -> Callable[..., R]:
        @functools.wraps(func)
        def f(self: 'Subject', *args: Any, **kwargs: Any) -> Any:
            if self.is_encrypted():
                raise TypeError(
                    f"{type(self).__name__} must be decrypted prior to calling "
                    f"{func.__name__}()."
                )
            return func(self, *args, **kwargs)
        return f

    @requires_decryption
    def get_masking_key(self) -> JSONWebKey:
        assert isinstance(self.dmk, JSONWebKey)
        return self.dmk

    def contribute_to_userinfo(self, userinfo: MutableMapping[str, Any]) -> None:
        userinfo['locale'] = self.locale
        userinfo['zoneinfo'] = self.zoneinfo

    def get_primary_key(self) -> SubjectKey:
        return SubjectKey(str(self.sub))

    def is_encrypted(self) -> bool:
        return any([
            isinstance(self.dek, EncryptionResult),
            isinstance(self.dmk, EncryptionResult),
        ])

    def update_from_oidc(self, token: OIDCToken) -> None:
        if token.locale:
            self.locale = token.locale
        if token.zoneinfo:
            self.zoneinfo = token.zoneinfo

    @requires_decryption
    async def decrypt(self, ct: EncryptionResult) -> Plaintext:
        assert not isinstance(self.dek, EncryptionResult)
        return await self.dek.decrypt(ct)

    @requires_decryption
    async def encrypt(self, value: bytes) -> EncryptionResult:
        assert not isinstance(self.dek, EncryptionResult)
        return await self.dek.encrypt(value)

    async def decrypt_keys(self, key: IDecrypter) -> None:
        """Decrypt the subjects' encryption keys using the
        storage encryption key.
        """
        if not isinstance(self.dek, JSONWebKey):
            assert isinstance(self.dek, EncryptionResult)
            assert isinstance(self.dmk, EncryptionResult)
            self.dek = JSONWebKey.model_validate_json(
                bytes.decode(bytes(await key.decrypt(self.dek)))
            )
            self.dmk = JSONWebKey.model_validate_json(
                bytes.decode(bytes(await self.dek.decrypt(self.dmk)))
            )

    async def encrypt_keys(self, key: IEncrypter) -> None:
        """Encrypt the subjects' encryption keys using the
        storage encryption key.
        """
        if isinstance(self.dek, JSONWebKey):
            assert not isinstance(self.dek, EncryptionResult)
            assert not isinstance(self.dmk, EncryptionResult)
            self.dmk = await self.dek.encrypt(
                str.encode(self.dmk.model_dump_json(exclude_none=True))
            )
            self.dek = await key.encrypt(
                str.encode(self.dek.model_dump_json(exclude_none=True))
            )

    @requires_decryption
    async def mask(self, value: bytes) -> str:
        assert isinstance(self.dmk, JSONWebKey)
        return b64encode(await self.dmk.sign(value), encoder=bytes.decode)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic
from aiopki.ext.jose import JOSEObject

from oauthx.lib.exceptions import InvalidRequest
from oauthx.lib.exceptions import UnknownClient
from oauthx.lib.protocols import IClient
from oauthx.lib.types import ClientAssertionType


class BaseGrant(pydantic.BaseModel):
    client_id: str | None = pydantic.Field(
        default=None,
        title="Client ID",
        description=(
            "This parameter is **required** if the client is not authenticating with "
            "the authorization server and the grant requires client identifiation, "
            "otherwise it **must** be omitted.\n\n"
            "If the client authenticates using an implementation of the RFC 7521 "
            "assertion framework, then the `client_id` parameter is unnecessary "
            "for client assertion authentication because the client is identified "
            "by the subject of the assertion.  If present, the value of the "
            "`client_id` parameter **must** identify the same client as is "
            "identified by the client assertion."
        )
    )

    client_secret: str | None = pydantic.Field(
        default=None,
        title="Client secret",
        description=(
            "This parameter is **required** if the client is authenticating using "
            "the `client_secret_post` method, otherwise is **must** be "
            "omitted."
        )
    )

    # RFC 7521 Assertion Framework for OAuth 2.0 Client
    # Authentication and Authorization Grants
    client_assertion_type: ClientAssertionType | None = pydantic.Field(
        default=None,
        title="Client assertion type",
        description=(
            "The format of the assertion as defined by the authorization server. "
            "The value is an absolute URI."
        )
    )

    client_assertion: JOSEObject | None = pydantic.Field(
        default=None,
        title='Client assertion',
        description=(
            "The assertion being used to authenticate the client. "
            "Specific serialization of the assertion is defined by "
            "profile documents for `client_assertion_type`."
        )
    )

    # RFC 8707 Resource Indicators for OAuth 2.0
    resource: str | None = pydantic.Field(
        default=None,
        description=(
            "Specifies the intended resource to use the issued access token and "
            "(optionally) ID token with."
        )
    )

    def model_post_init(self, _: Any):
        if self.client_secret and (self.client_assertion or self.client_assertion_type):
            raise InvalidRequest(
                "The request must not utilize more than one "
                "mechanism for authenticating the client"
            )
        if (self.client_assertion_type or self.client_assertion)\
        and not all([self.client_assertion_type, self.client_assertion]):
            raise UnknownClient(
                "Provide both the `client_assertion_type` and "
                "`client_assertion` parameters."
            )
        if self.client_secret and not self.client_id:
            raise UnknownClient(
                "The `client_secret` parameter is required."
            )

    def has_credentials(self) -> bool:
        return any([
            self.client_secret is not None,
            self.client_assertion_type is not None
        ])

    def must_authenticate(self) -> bool:
        """Return a boolean if the client must always authenticate when
        using this grant.
        """
        return True

    def must_identify(self) -> bool:
        """Return a boolean if the client must identify when
        using this grant.
        """
        return True

    def requires_offline_access(self) -> bool:
        """Return a boolean indicating if the grant requires the
        ``offline_access`` scope.
        """
        return False

    def verify(self, client: IClient):
        """Verifies that the grant is valid for the given client."""
        pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Protocol
from typing import TypeVar

from canonical import ResourceIdentifier


T = TypeVar('T')


class IClientAuthorizationState(Protocol):
    __module__: str = 'oauthx.types'
    client_id: str

    class Key(ResourceIdentifier[str, 'IClientAuthorizationState']):
        state: str
        openapi_example: str = 'EwzP5eWNArizDvJdO95fuyoKuLDpoYmOx2ZGuqudqWTZy7QlSxxkXyx2fNwhjjI8'
        openapi_title: str = 'State'

        def __init__(self, state: str):
            self.state = state

        def cast(self) -> str:
            return str(self.state)

        def __str__(self) -> str:
            return self.state
        
        def __eq__(self, key: object) -> bool:
            return isinstance(key, type(self)) and key.state == self.state

        def __hash__(self) -> int:
            return hash(self.state)

    @property
    def key(self) -> Key:
        ...

    @property
    def pk(self) -> Key:
        return self.Key(self.client_id)

    def annotate(self, key: str, value: Any) -> None: ...
    def annotation(self, key: str, decoder: Callable[[Any], T] = lambda x: x) -> T | None: ...
    def get_authorize_url(self) -> str: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import itertools
from typing import AsyncIterable
from typing import Iterable

from oauthx.lib.types import OIDCIssuerIdentifier
from .claim import Claim


class ClaimSet:
    _known: set[tuple[str, str]]

    @classmethod
    async def fromquery(cls, provider: OIDCIssuerIdentifier, query: AsyncIterable[Claim]):
        return  cls(provider, [claim async for claim in query])

    def __init__(self, provider: OIDCIssuerIdentifier, claims: Iterable[Claim]):
        self.userinfo = {}
        self.provider = provider
        self.claims = list(claims)
        self._known = {(str(x.root.kind), str(x.root.issuer)) for x in claims}

    def is_known(self, claim: Claim) -> bool:
        return (str(claim.root.kind), str(claim.root.issuer)) in self._known

    def diff(self, claims: Iterable[Claim]) -> list[Claim]:
        """Return the :class:`Claim` objects from iterable `claims`
        that have changed with respect to the claims contained in the
        :class:`ClaimSet`.
        """
        claims = [x for x in claims if x.root.provider == self.provider]
        changed: list[Claim] = [x for x in claims if not self.is_known(x)]
        for old, new in itertools.product(self.claims, claims):
            if new in changed:
                continue
            if old.root.kind != new.root.kind:
                continue
            if (old.root.issuer, old.root.masked) == (new.root.issuer, new.root.masked):
                continue
            changed.append(new)
        return changed

    def __bool__(self) -> bool:
        return bool(self.claims)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .masked import Masked


__all__: list[str] = [
    'Unmasked'
]


class Unmasked(Masked):
    __module__: str = 'oauthx.types'

    def __init__(self, masked: str = ''):
        super().__init__(masked)

    def __repr__(self) -> str:
        return f'<Unmasked>'

    def __str__(self) -> str:
        raise NotImplementedError

    def __hash__(self) -> int:
        raise NotImplementedError

    def __eq__(self, key: object) -> bool:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from ..config import Config


__all__: list[str] = ['CurrentConfig']


def get(request: fastapi.Request) -> Config:
    return request.config # type: ignore



CurrentConfig: Config = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from .baseresponsemode import BaseResponseMode


class FragmentResponseMode(BaseResponseMode):
    response_mode: Literal['fragment', 'fragment.jwt']
    response_type: Literal['id_token', 'code token', 'id_token token', 'code id_token token']

    @pydantic.field_validator('response_mode', mode='before')
    def preprocess_response_mode(cls, value: str | None) -> str:
        if value is None:
            value = 'fragment'
        return value



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from aiopki.utils import b64decode_int
from canonical import ResourceIdentifier

from oauthx.lib.types import RequestURI

from .iauthorizationrequest import IAuthorizationRequest


class AuthorizationRequestKey(ResourceIdentifier[int, IAuthorizationRequest]):
    __module__: str = 'oauthx.server.types'

    @classmethod
    def fromuri(cls, request_uri: RequestURI):
        assert request_uri.id is not None
        return cls(b64decode_int(request_uri.id))

    def __init__(self, pk: int):
        self.__pk = pk

    def cast(self) -> int:
        return self.__pk

    def __int__(self) -> int:
        return int(self.__pk)

    def __str__(self) -> str:
        return str(self.__pk)

    def __hash__(self) -> int:
        return hash(self.__pk)

    def __repr__(self) -> str:
        return f'<AuthorizationRequestKey: {str(self)}>'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.server import models
from oauthx.server.params import ClientAuthorizationState
from oauthx.server.params import Storage
from oauthx.server.types import AuthorizationRequestKey
from oauthx.server.types import StopSnooping


__all__: list[str] = [
    'AuthorizationRequest'
]


async def get(
    storage: Storage,
    state: ClientAuthorizationState
) -> models.AuthorizationRequest:
    if state is None:
        raise StopSnooping
    request_id = state.annotation('request')
    if request_id is None:
        raise StopSnooping
    return cast(
        models.AuthorizationRequest,
        await storage.get(AuthorizationRequestKey(int(request_id)))
    )


AuthorizationRequest: TypeAlias = Annotated[
    models.AuthorizationRequest,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Union
from typing import TypeAlias

import fastapi
import pydantic

from oauthx.lib.types import GrantType
from oauthx.lib.types import NullRedirectURI
from oauthx.lib.types import ResponseType
from .fragmentresponsemode import FragmentResponseMode
from .queryresponsemode import QueryResponseMode


__all__: list[str] = ['ResponseMode']


ResponseModeType: TypeAlias = Union[
    FragmentResponseMode,
    QueryResponseMode
]


class ResponseMode(pydantic.RootModel[ResponseModeType]):

    @property
    def grant_type(self) -> GrantType | None:
        return self.root.grant_type

    @property
    def response_type(self) -> ResponseType:
        return self.root.response_type # type: ignore

    def can_redirect(self) -> bool:
        return not isinstance(self.root.redirect_uri, NullRedirectURI)

    def grants(self) -> GrantType | None:
        return self.grant_type

    def get_token_types(self) -> set[str]:
        return self.root.get_token_types()

    async def deny(self) -> str:
        return await self.root.deny()

    async def error(
        self,
        *,
        error: str,
        error_description: str | None = None,
        error_uri: str | None = None
    ) -> str:
        return await self.root.error(
            error=error,
            error_description=error_description,
            error_uri=error_uri,
        )

    async def redirect(self, **kwargs: Any) -> fastapi.responses.RedirectResponse:
        return await self.root.redirect(**kwargs)

    def __str__(self) -> str:
        return self.root.response_mode


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.exceptions import InvalidRequest
from .authorizationrequest import AuthorizationRequest
from .client import Client
from .query import STATE


__all__: list[str] = [
    'State'
]


async def get(
    client: Client,
    params: AuthorizationRequest,
    state: str | None = STATE,
) -> str | None:
    if params and state:
        raise InvalidRequest(
            "The state parameter must not be included if the "
            "request or request_uri parameters are present."
        )
    if params:
        state = params.state
    return state


State: TypeAlias = Annotated[
    str | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from typing import Any

from oauthx.lib.models import ClientAuthorizationState
from oauthx.lib.models import ClientAuthorizationStateKey
from oauthx.lib.types import MaskedPrincipal
from oauthx.server.models import Authorization
from oauthx.server.models import AuthorizationRequest
from oauthx.server.models import Client
from oauthx.server.models import Principal
from oauthx.server.models import ResourceOwner
from oauthx.server.models import Subject
from oauthx.server.models import SubjectKey
from oauthx.server.params import ContentEncryptionKey
from oauthx.server.protocols import IResourceOwner
from oauthx.server.protocols import ISubject
from oauthx.server.types import AuthorizationKey
from ..basestorage import BaseStorage


class Storage(BaseStorage):
    __module__: str = 'oauthx.server.ref'
    objects: dict[Any, Any] = {}
    key: ContentEncryptionKey

    def setup(
        self,
        *,
        key: ContentEncryptionKey,
        **kwargs: Any) -> None:
        self.key = key
        self.objects = Storage.objects

    async def allocate_identifier(self, cls: str | type) -> int:
        return secrets.choice(range(10000000, 99999999))

    async def delete(self, obj: Any) -> None:
        self.objects.pop(obj.pk, None)

    async def get_authorization(self, key: AuthorizationKey) -> Authorization | None:
        return self.objects[key]

    async def get_authorization_state(self, key: ClientAuthorizationStateKey) -> Any:
        return self.objects[key]

    async def get_client(self, key: Client.Key) -> Client | None:
        return Client.model_validate(self.objects[f'clients/{key.client_id}'])

    async def get_principal(self, masked: MaskedPrincipal) -> Principal | None:
        return self.objects[masked]

    async def get_resource_owner(self, key: IResourceOwner.Key) -> ResourceOwner | None:
        return self.objects[key]

    async def get_subject_by_pk(self, pk: SubjectKey) -> ISubject | None:
        subject = self.objects.get(pk)
        if subject is not None:
            await subject.decrypt_keys(self.key)
        return subject

    async def get_subject_by_principal(self, masked: MaskedPrincipal) -> ISubject | None:
        principal = self.objects.get(masked)
        if principal is None:
            return None
        subject = self.objects[SubjectKey(str(principal.owner))]
        if subject is not None:
            await subject.decrypt_keys(self.key)
        return subject

    async def persist_authorization(self, authorization: Authorization) -> None:
        self.objects[authorization.pk] = authorization

    async def persist_authorization_request(self, request: AuthorizationRequest) -> None:
        self.objects[request.pk] = request

    async def persist_authorization_state(self, state: ClientAuthorizationState) -> None:
        self.objects[state.pk] = state

    async def persist_client(self, client: Client) -> None:
        self.objects[f'clients/{client.root.client_id}'] = client

    async def persist_principal(self, principal: Principal) -> None:
        self.objects[principal.masked] = principal

    async def persist_resource_owner(self, owner: ResourceOwner) -> None:
        self.objects[owner.pk] = owner

    async def persist_subject(self, subject: Subject) -> None:
        await subject.encrypt_keys(self.key)
        self.objects[subject.get_primary_key()] = subject


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import warnings
from typing import Annotated
from typing import TypeAlias

import fastapi
from aiopki.ext import jose

from oauthx.lib.params import ClientAuthorizationState
from .obtainedgrant import ObtainedGrant
from .requestingclient import RequestingClient
from .upstreamprovider import UpstreamProvider


__all__: list[str] = ['OIDCToken']


def get(
    client: RequestingClient,
    state: ClientAuthorizationState,
    grant: ObtainedGrant,
    provider: UpstreamProvider
) -> jose.OIDCToken:
    warnings.warn("OIDC ID Token is not verified.")
    if grant.id_token is None:
        raise NotImplementedError
    token = grant.id_token.payload(jose.OIDCToken.model_validate)
    return provider.process_oidc_token(client, state, token)


OIDCToken: TypeAlias = Annotated[jose.OIDCToken, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
import fastapi.params

from oauthx.lib.exceptions import InsufficientScope
from .accesstoken import AccessToken


__all__: list[str] = ['RequiredScope']


def RequiredScope(scope: set[str]) -> fastapi.params.Depends:
    def f(at: AccessToken) -> None:
        if not at.validate_scope(scope):
            raise InsufficientScope
    return fastapi.Depends(f)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.types import GrantType as GrantType_
from oauthx.server.types import UnauthorizedClient
from .client import Client
from .query import GRANT_TYPE


__all__: list[str] = [
    'GrantType'
]


async def get(
    client: Client,
    grant_type: GrantType_ = GRANT_TYPE
) -> GrantType_:
    if client is not None and not client.can_grant(grant_type):
        raise UnauthorizedClient
    return grant_type


GrantType: TypeAlias = Annotated[
    GrantType_,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
import aiopki.types

from oauthx.server.request import Request


__all__: list[str] = ['SessionSigner']


def get(request: Request) -> aiopki.types.ICryptoKey:
    return request.signer


SessionSigner: TypeAlias =  Annotated[aiopki.types.ICryptoKey, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any
from typing import Literal

import fastapi
from aiopki.ext.jose import OIDCToken

from oauthx.lib.params import Logger
from oauthx.lib.types import OIDCIssuerIdentifier
from oauthx.server.protocols import IAuthorizationServer
from oauthx.server.protocols import ISubject
from .params import ContentEncryptionKey
from .params import ObjectFactory
from .params import Storage
from .params import SubjectLogger
from .subjectresolver import SubjectResolver


class AuthorizationServer(IAuthorizationServer):
    __module__: str = 'oauthx.server'
    factory: ObjectFactory
    non_personal_claims: set[str] = {
        'aud', 'exp', 'iss', 'nbf', 'at_hash', 'c_hash',
        'nonce', 'iat', 'azp', 'acr', 'amr'
    }
    storage: Storage
    subjects: SubjectResolver

    def __init__(
        self,
        logger: Logger,
        storage: Storage,
        encryption_key: ContentEncryptionKey,
        factory: ObjectFactory,
        subject_publisher: SubjectLogger,
        subjects: SubjectResolver = fastapi.Depends(SubjectResolver)
    ):
        self.encryption_key = encryption_key
        self.factory = factory
        self.logger = logger
        self.storage = storage
        self.subjects = subjects
        self.subject_publisher = subject_publisher

    async def onboard_oidc(
        self,
        token: OIDCToken,
        use: Literal['personal', 'institutional'],
        request_id: int | None = None,
        client_id: str | None = None
    ) -> tuple[ISubject, bool]:
        created = False
        issuer = OIDCIssuerIdentifier(token.iss)
        register: set[Any] = set()
        subject = None

        # Try to get subject by the identifier returned by the identity provider.
        # If the identifier yields no subject, then this provider was not used
        # before to login. In that case, it is safe to lookup the subject by
        # its email address and associate the identifier. If the sub claim
        # does resolve to a principal, the email address MUST match the
        # known email address or be None.
        email = await self.subjects.resolve_principal(token.email)
        subject_id = await self.subjects.resolve_principal(token.subject_id)
        if subject_id and email:
            # If there is an email principal, then the sub claims must match.
            if not email.is_owned_by(subject_id.owner):
                # We have a problem.
                self.logger.warning("OIDC ID Token resolved to multiple principals.")
                raise self.AmbiguousPrincipal
            subject = await self.storage.get(email.owner)
            if subject is None:
                self.logger.critical(
                    "Could not retrieve Subject with the given Principal "
                    "(kind: %s, mask: %s)",
                    email.kind, email.masked
                )
                raise self.OrphanedPrincipal
        elif not any([subject_id, email]):
            # This is a new user and we can safely created it.
            created = True
            subject = await self.factory.subject(use)
            subject.update_from_oidc(token)
            register.update({token.subject_id, token.email, token.phone_number})
        elif subject_id and not email:
            subject = await self.storage.get(subject_id.owner)
            if subject is None:
                self.logger.critical(
                    "Could not retrieve Subject with the given Principal "
                    "(kind: %s, mask: %s)",
                    subject_id.kind, subject_id.masked
                )
                raise self.OrphanedPrincipal
            register.update({token.email, token.phone_number})
        elif email and not subject_id:
            subject = await self.storage.get(email.owner)
            if subject is None:
                self.logger.critical(
                    "Could not retrieve Subject with the given Principal "
                    "(kind: %s, mask: %s)",
                    email.kind, email.masked
                )
                raise self.OrphanedPrincipal
            register.update({token.subject_id, token.phone_number})
        else:
            raise NotImplementedError
        register.remove(None)
        assert subject is not None
        assert None not in register
        if created:
            self.logger.info(
                "Registered subject from OpenID Connect "
                "(iss: %s, sub: %s)",
                token.iss, subject.get_primary_key()
            )
            await subject.encrypt_keys(self.encryption_key)
            await self.storage.persist(subject)
        await subject.decrypt_keys(self.encryption_key) # type: ignore
        if token.subject_id in register:
            await self.register_principal(
                issuer=OIDCIssuerIdentifier(token.iss),
                subject=subject,
                value=token.subject_id,
                created=created,
                verified=True
            )
        if token.email in register:
            await self.register_principal(
                issuer=OIDCIssuerIdentifier(token.iss),
                subject=subject,
                value=token.email,
                created=created,
                verified=token.email_verified
            )
        if token.phone_number in register:
            await self.register_principal(
                issuer=issuer,
                subject=subject,
                value=token.email,
                created=created,
                verified=token.phone_number_verified
            )

        receipt = await self.factory.receipt(
            provider=token.iss,
            purpose='LOGIN',
            sub=int(subject.get_primary_key()), # type: ignore
            claims={k for k in token.model_dump(
                exclude=self.non_personal_claims,
                exclude_none=True
            )},
            request_id=request_id,
            client_id=client_id
        )
        async with receipt.transaction(subject, self.storage, self.subject_publisher) as tx:
            tx.add('birthdate', token.gender, issuer='self')
            tx.add('gender', token.gender, issuer='self')
            tx.add('email', token.email,
             issuer='self' if not token.email_verified else issuer)
            tx.add('phone_number', token.phone_number,
             issuer='self' if not token.phone_number_verified else issuer)
            tx.add('name', token.name, issuer='self')
            tx.add('given_name', token.given_name, issuer='self')
            tx.add('middle_name', token.middle_name, issuer='self')
            tx.add('family_name', token.family_name, issuer='self')
            tx.add('nickname', token.nickname, issuer='self')

        return subject, created

    async def register_principal(
        self,
        *,
        issuer: OIDCIssuerIdentifier,
        subject: ISubject,
        value: Any,
        created: bool = False,
        verified: bool = False
    ) -> None:
        now = datetime.datetime.now(datetime.timezone.utc)
        new = await self.factory.principal(
            subject=subject,
            issuer=issuer,
            owner=subject.get_primary_key(), # type: ignore
            now=now,
            value=value,
            verified=verified
        )
        old = await self.storage.get(new.masked)
        if not old or not old.is_verified():
            await new.encrypt(subject)
            await self.storage.persist(new)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from oauthx.lib.exceptions import InteractionException


class InvalidClient(InteractionException):
    __module__: str = 'oauthx.server.types'
    error: str = 'invalid_client'
    error_description: str = (
        'Client authentication failed (e.g., unknown client, '
        'no client authentication included, or unsupported '
        'authentication method)'
    )
    status_code = 401
    template_name = 'oauthx/errors/authorize.invalid_client.html.j2'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Awaitable
from typing import Callable
from typing import TypeAlias

from oauthx.lib import AuthorizationRequest
from oauthx.lib import ClientAuthorizationState
from oauthx.lib import Error
from oauthx.lib import RedirectionParameters


AuthorizeReturn: TypeAlias = tuple[AuthorizationRequest, Error | RedirectionParameters, ClientAuthorizationState]

AuthorizeFactory: TypeAlias = Callable[..., Awaitable[AuthorizeReturn]]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic
from canonical import PythonSymbol

from oauthx.server.protocols import IRequestSubject


class LoginHandler(pydantic.BaseModel):
    name: str
    path: str
    impl: PythonSymbol[Any]


class AuthenticationConfig(pydantic.BaseModel):
    impl: PythonSymbol[type[IRequestSubject]]
    endpoint: str = '/login'
    handlers: list[LoginHandler]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

import fastapi

from oauthx.lib.protocols import IClient
from oauthx.lib.protocols import IStorage


class IAuthorizationRequest(Protocol):
    __module__: str = 'oauthx.server.types'
    
    @property
    def pk(self) -> Any: ...
    
    def add_to_template_context(self, request: fastapi.Request, context: dict[str, Any]) -> None: ...
    async def get_client(self, storage: IStorage) -> IClient: ...
    async def get_context(self, storage: IStorage) -> Any: ...
    async def deny(self, client: IClient, exception: bool = False) -> str: ...
    async def redirect(self, client: IClient, **params: Any) -> fastapi.Response: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import itertools
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.server.models import AuthorizationRequest
from oauthx.server.protocols import IClient
from oauthx.server.request import Request
from .currentsubject import CurrentSubject
from .objectfactory import ObjectFactory
from .storage import Storage


__all__: list[str] = ['PluginRunner']


class _PluginRunner:
    plugins: list

    def __init__(
        self,
        request: Request,
        factory: ObjectFactory,
        storage: Storage,
        subject: CurrentSubject
    ):
        self.plugins = [ # type: ignore
            cls(request, factory, storage, subject)
            for cls in request.plugins  # type: ignore
        ]
        self.request = request
        self.storage = storage
        self.subject = subject

    async def validate_scope(
        self,
        client: IClient,
        request: AuthorizationRequest,
        scope: set[str]
    ) -> fastapi.Response | None:
        """Validates the requested scope when authorizing a request
        or issueing a token.
        """
        response = None
        for plugin, name in itertools.product(self.plugins, sorted(scope)):
            if not plugin.handles_scope(name):
                continue
            name, params = await plugin.resolve_scope(client, request, name)
            if not name:
                continue
            url = self.request.url_for(name, pk=request.id, **params)
            response = fastapi.responses.RedirectResponse(
                status_code=302,
                url=url
            )
            break
        return response


PluginRunner: TypeAlias = Annotated[_PluginRunner, fastapi.Depends(_PluginRunner)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import secrets
from typing import Any
from typing import Literal

import aiopki
import fastapi
import pydantic
from aiopki.ext import jose
from aiopki.lib import JSONWebKeySet

from oauthx.lib.types import GrantType
from .baseresponsemode import BaseResponseMode


class QueryResponseMode(BaseResponseMode):
    response_mode: Literal['query', 'jwt', 'query.jwt']
    response_type: Literal['code', 'code id_token', 'none']

    @property
    def grant_type(self) -> GrantType | None:
        return 'authorization_code' if self.response_type != 'none' else None

    @pydantic.field_validator('response_mode', mode='before')
    def preprocess_response_mode(cls, value: str | None) -> str:
        if value is None:
            value = 'query'
        return value

    def get_responses(self) -> set[str]:
        return {x for x in str.split(self.response_type, ' ')}

    async def deny(self) -> str:
        return self.get_redirect_uri().redirect(
            mode='query',
            error='access_denied',
            state=self.state,
            iss=self.iss
        )

    async def error(
        self,
        *,
        error: str,
        error_description: str | None = None,
        error_uri: str | None = None
    ) -> str:
        redirect_uri = self.get_redirect_uri()
        return redirect_uri.redirect(
            False,
            mode='query',
            error=error,
            error_description=error_description,
            error_uri=error_uri,
            state=self.state,
            iss=self.iss
        )

    async def redirect(self, signer: aiopki.CryptoKeyType, **kwargs: Any) -> fastapi.responses.RedirectResponse:
        if self.response_type == 'none':
            kwargs = {}
        kwargs.update({
            'iss': self.iss,
            'state': self.state
        })
        if self.response_mode in {'query.jwt', 'jwt'}:
            kwargs = {
                'response': await self.sign(signer, **kwargs)
            }
        return fastapi.responses.RedirectResponse(
            status_code=302,
            url=self.get_redirect_uri().redirect(False, mode='query', **kwargs)
        )

    async def sign(
        self,
        signer: aiopki.CryptoKeyType,
        **claims: Any
    ) -> str:
        now = datetime.datetime.now(datetime.timezone.utc)
        jws = jose.jws({
            **claims,
            'jti': secrets.token_urlsafe(32),
            'aud': self.client.id,
            'exp': now + datetime.timedelta(seconds=60),
            'nbf': now,
            'iat': now,
        })
        await jws.sign(signer.default_algorithm(), signer)
        if self.client.wants_encryption():
            assert isinstance(self.client.credential, JSONWebKeySet)
            assert self.client.credential.can_encrypt()
            await jws.encrypt(
                encrypter=self.client.credential
            )
        return jws.encode(encoder=bytes.decode, compact=True)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .error import Error


class FatalError(Error):
    """An implementation of :class:`~headless.ext.oicd.types.Error` that
    represent an exception that fatally exits the authorization flow.
    """
    __module__: str = 'headless.ext.oidc.types'

    def fatal(self) -> bool:
        return True


class InvalidScope(FatalError, ValueError):
    __module__: str = 'headless.ext.oidc.types'

    def __init__(self, description: str, url: str | None = None):
        FatalError.__init__(
            self,
            error='invalid_scope',
            error_description=description,
            error_uri=url
        )


class UnauthorizedClient(FatalError):
    __module__: str = 'headless.ext.oidc.types'

    def __init__(self, description: str, url: str | None = None):
        super().__init__(
            error='unauthorized_client',
            error_description=description,
            error_uri=url
        )


class UnsupportedResponseType(FatalError):
    __module__: str = 'headless.ext.oidc.types'

    def __init__(self, description: str, url: str | None = None):
        super().__init__(
            error='unsupported_response_type',
            error_description=description,
            error_uri=url
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import secrets
import warnings
from typing import Any
from typing import Iterable
from typing import TypeVar

import httpx
import pydantic
from aiopki.ext.jose import JWA
from aiopki.lib import JSONWebKeySet

from oauthx.lib import utils
from oauthx.lib import AuthorizationResponse
from oauthx.lib import ClientAuthorizationState
from oauthx.lib import Error
from oauthx.lib import ObtainedGrant
from oauthx.lib import Grant
from oauthx.lib import Provider
from oauthx.lib import RedirectionParameters
from oauthx.lib.types import ClientCredentialType
from oauthx.lib.types import PrompType
from oauthx.lib.types import ResponseModeType
from oauthx.lib.types import ResponseType


T = TypeVar('T', bound=ClientAuthorizationState)


class Client(pydantic.BaseModel):
    auth_method: str | None = None
    client_id: str
    credential: ClientCredentialType | None = None
    port: int | None = None
    provider: Provider
    response_mode: ResponseModeType = 'query'
    response_type: ResponseType = 'code'

    # JWT Secured Authorization Response Mode for OAuth 2.0 (JARM)
    authorization_signed_response_alg: JWA = pydantic.Field(
        default=JWA.rs256
    )

    model_config = {
        'arbitrary_types_allowed': True,
    }

    @pydantic.field_validator('provider', mode='before')
    def preprocess_provider(
        cls,
        value: Provider | dict[str, str] | str
    ) -> Provider:
        if isinstance(value, str):
            value = {'issuer': value}
        if isinstance(value, dict):
            if not value.get('issuer'):
                raise ValueError("The `issuer` parameter is required.")
            value = Provider.model_validate({
                'iss': value.get('issuer'),
                **value
            })
        return value

    def can_encrypt(self) -> bool:
        return any([
            isinstance(self.credential, JSONWebKeySet) and self.credential.can_encrypt()
        ])

    def get_local_port(self) -> int:
        return self.port or utils.random_port()

    def is_confidential(self) -> bool:
        return isinstance(self.credential, JSONWebKeySet)

    async def authenticate(self, grant: Grant) -> bool:
        grant.root.client_id = self.client_id
        if isinstance(self.credential, str):
            grant.root.client_secret = self.credential
        return True

    @utils.http
    async def authorize(
        self,
        http: httpx.AsyncClient,
        *,
        redirect_uri: str | None = None,
        scope: Iterable[str] | None = None,
        response_type: ResponseType | None = None,
        state_class: type[T] = ClientAuthorizationState,
        prompt: Iterable[PrompType] = [],
        annotations: dict[str, str] | None = None,
        **params: Any
    ) -> T:
        """Create an authorization request and return the URL to
        which a user-agent must be redirectred.
        """
        await self.provider.discover(http)
        scope = scope or set()
        params = {
            **params,
            'client_id': self.client_id,
            'redirect_uri': redirect_uri,
            'scope': str.join(' ', sorted(scope)),
            'response_type': response_type or self.response_type,
            'response_mode': self.response_mode,
            'state': secrets.token_urlsafe(32)
        }
        if 'openid' in scope:
            warnings.warn('OpenID Connect is partially implemented')
            params.update({
                'nonce': secrets.token_urlsafe(32),
                'prompt': str.join(' ', set(prompt))
            })
        url = self.provider.authorize(**{k: v for k, v in params.items() if v})
        return state_class.model_validate({
            'authorize_url': url,
            'client_id': self.client_id,
            'params': params,
            'annotations': annotations or {}
        })

    async def authorization_code( # type: ignore
        self,
        params: AuthorizationResponse | Error | RedirectionParameters,
        state: ClientAuthorizationState,
        http: httpx.AsyncClient | None = None
    ) -> ObtainedGrant:
        """Use the query parameters that the authorization server
        supplied to the client redirection endpoint to obtain
        an access token.
        """
        if not isinstance(params, (Error, RedirectionParameters)):
            params = await self.on_redirected(state, params)
        params.raise_for_status()
        assert isinstance(params, RedirectionParameters), type(params)
        grant = Grant.model_validate({
            'grant_type': 'authorization_code',
            'code': params.code,
            'redirect_uri': state.redirect_uri
        })
        obtained = await self.grant(grant, http=http)
        if obtained.id_token is not None:
            if obtained.id_token.is_encrypted():
                raise NotImplementedError
            await self.provider.verify_id_token(obtained.id_token)
        return obtained

    async def discover(self, http: httpx.AsyncClient) -> None:
        await self.provider.discover(http)

    @functools.singledispatchmethod
    async def grant(
        self,
        params: AuthorizationResponse,
        *args: Any,
        **kwargs: Any
    ) -> ObtainedGrant:
        raise NotImplementedError(params)

    @grant.register
    async def _(
        self,
        grant: Grant,
        authenticate: bool = True,
        http: httpx.AsyncClient | None = None
    ) -> ObtainedGrant:
        if authenticate:
            await self.authenticate(grant)
        return await self.provider.grant(grant, http=http)

    async def on_redirected(
        self,
        state: ClientAuthorizationState,
        params: dict[str, Any] | AuthorizationResponse | Error | RedirectionParameters
    ) -> Error | RedirectionParameters:
        """Processes the parameters that were supplied by the authorization
        server to the redirection endpoint.
        """
        await self.provider
        if not isinstance(params, (Error, RedirectionParameters)):
            if not isinstance(params, AuthorizationResponse):
                params = AuthorizationResponse.model_validate(params)
            await params.decode(
                issuer=self.provider.issuer,
                client_id=self.client_id,
                verifier=self.provider.signing_keys,
                decrypter=self.credential
            )
            assert isinstance(params.root, (Error, RedirectionParameters))
            params = params.root

        if isinstance(params, RedirectionParameters):
            await self.provider.verify(params)
            await state.verify(params)
        return params

    async def userinfo(self, access_token: str, http: httpx.AsyncClient | None = None):
        return await self.provider.userinfo(access_token, http=http)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import functools
from typing import Any
from typing import Callable
from typing import Literal
from typing import MutableMapping
from typing import TypeVar

import pydantic
from aiopki.types import EncryptionResult
from aiopki.types import ISigner
from aiopki.utils import b64encode

from oauthx.lib.types import Masked
from oauthx.lib.types import Unmasked
from oauthx.lib.types import OIDCIssuerIdentifier
from oauthx.server.protocols import ISubject


R = TypeVar('R')
T = TypeVar('T', bound='BaseClaim')


class BaseClaim(pydantic.BaseModel):
    """The base class for all claim implementations. A :term:`Claim`
    is an assertion about a :term:`Subject`; it represents an attribute
    of the identified user.
    """
    id: int = pydantic.Field(
        default=...
    )

    ial: int = pydantic.Field(
        default=0
    )

    issuer: Literal['self'] | OIDCIssuerIdentifier = pydantic.Field(
        default='self'
    )

    kind: Any = pydantic.Field(
        default=...
    )

    masked: Masked = pydantic.Field(
        default_factory=Unmasked
    )

    obtained: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    provider: Literal['self'] | OIDCIssuerIdentifier = pydantic.Field(
        default=...
    )

    sub: int = pydantic.Field(
        default=...
    )

    receipt_id: int = pydantic.Field(
        default=...
    )

    value: Any

    @property
    def key(self) -> tuple[str, str, str]:
        return (
            str(self.issuer),
            str(self.provider),
            str(self.kind)
        )

    @staticmethod
    def requires_decryption(func: Callable[..., R]) -> Callable[..., R]:
        @functools.wraps(func)
        def f(self: 'BaseClaim', *args: Any, **kwargs: Any) -> Any:
            if self.is_encrypted():
                raise TypeError(
                    f"{type(self).__name__} must be decrypted prior to calling "
                    f"{func.__name__}()."
                )
            return func(self, *args, **kwargs)
        return f

    @pydantic.field_serializer('value', mode='plain', when_used='always')
    def dump_model(self, value: EncryptionResult, _: Any) -> dict[str, Any]:
        return value.root.model_dump(mode='json')

    def contribute_to_userinfo(self, userinfo: MutableMapping[str, Any]) -> None:
        userinfo[self.kind] = self.get_value()

    def decode_plaintext(self, pt: bytes) -> Any:
        obj = self.model_validate({
            **self.model_dump(exclude={'value'}),
            'value': bytes.decode(pt, 'utf-8')
        })
        return obj.value

    def get_encryption_input(self) -> bytes:
        raise NotImplementedError

    def get_masking_input(self) -> bytes:
        raise NotImplementedError

    @requires_decryption
    def get_value(self) -> Any:
        return self.value

    def is_encrypted(self) -> bool:
        return isinstance(self.value, EncryptionResult)

    def is_masked(self):
        return not isinstance(self.masked, Unmasked)

    def set_mask(self, masked: bytes):
        self.masked = Masked(b64encode(masked, encoder=bytes.decode))

    async def decrypt(self, subject: ISubject) -> None:
        if isinstance(self.value, EncryptionResult):
            pt = await subject.decrypt(self.value)
            self.value = self.decode_plaintext(bytes(pt))

    async def encrypt(self, subject: ISubject) -> None:
        """Encrypt the claim using the owners' private key."""
        if not isinstance(self.value, EncryptionResult):
            self.value = await subject.encrypt(self.get_encryption_input())

    @requires_decryption
    async def mask(self, key: ISigner) -> None:
        self.set_mask(await key.sign(self.get_masking_input()))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os

import httpx
import pytest

import oauthx
from .types import AuthorizeFactory


@pytest.mark.asyncio
async def test_client_id_must_match(
    http: httpx.AsyncClient,
    client: oauthx.Client,
    authorize_request: AuthorizeFactory
):
    _, params, state = await authorize_request(response_type='code')
    try:
        client.client_id = bytes.hex(os.urandom(16))
        await client.authorization_code(params, state, http=http)
        assert False
    except oauthx.Error as exc:
        assert exc.error == 'invalid_client'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import functools
from typing import Any
from typing import AsyncIterable
from typing import Callable
from typing import Iterable
from typing import Mapping
from typing import TypeVar

import pydantic
from canonical import ResourceIdentifier

from oauthx.server.params import ContentEncryptionKey
from .params import DatastoreClient
from .types import IDatastoreKey
from .types import IDatastoreEntity


T = TypeVar('T', bound=pydantic.BaseModel)


class DatastoreStorage:

    @staticmethod
    async def run_in_executor(
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any
    ) -> Any:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, func, *args, **kwargs)

    def __init__( # type: ignore
        self,
        *,
        client: DatastoreClient, # type: ignore
        key: ContentEncryptionKey
    ) -> None:
        self.client = client
        self.key = key

    async def allocate_identifier(self, cls: type | str) -> int:
        if not isinstance(cls, str):
            cls = cls.__name__
        base = self.entity_key(kind=cls)
        result = await self.run_in_executor(
            functools.partial( # type: ignore
                self.client.allocate_ids, # type: ignore
                incomplete_key=base,
                num_ids=1
            )
        )
        return [x for x in result][0].id

    async def delete(self, obj: Any) -> None:
        key = self.entity_key(type(obj), obj.pk)
        await self.run_in_executor(
            functools.partial(
                self.client.delete, # type: ignore
                key=key
            )
        )

    def entity_factory(self, key: IDatastoreKey, obj: pydantic.BaseModel) -> IDatastoreEntity:
        entity = self.client.entity(key) # type: ignore
        attrs = obj.model_dump(mode='json')
        if attrs:
            entity.update(attrs) # type: ignore
        return entity # type: ignore

    def entity_key(
        self,
        kind: str | type,
        identifier: int | str | ResourceIdentifier[int | str, Any] | None = None,
        parent: IDatastoreKey | None = None
    ) -> IDatastoreKey:
        if not isinstance(kind, str):
            kind = kind.__name__
        if isinstance(identifier, ResourceIdentifier):
            identifier = identifier.cast() # type: ignore
        args: list[Any] = [kind]
        if identifier is not None:
            args.append(identifier)
        return self.client.key(*args, parent=parent) # type: ignore

    def model_factory(self, entity: Mapping[str, Any], model: type[T]) -> T:
        return model.model_validate(dict(entity))

    async def find(
        self,
        kind: type[T],
        filters: Iterable[tuple[str, str, int | str]],
        sort: Iterable[str] | None = None
    ) -> AsyncIterable[T]:
        q = self.client.query(kind=kind.__name__) # type: ignore
        if sort is not None:
            q.order = sort
        for filter in filters:
            q.add_filter(*filter) # type: ignore
        results = await self.run_in_executor(
            functools.partial(q.fetch) # type: ignore
        )
        for entity in results:
            yield kind.model_validate(dict(entity))

    async def get_entity_by_key(self, key: IDatastoreKey) -> IDatastoreEntity | None:
        return await self.run_in_executor(
            functools.partial(
                self.client.get, # type: ignore
                key=key
            )
        )

    async def get_model_by_key(self, cls: type[T], pk: int | str | ResourceIdentifier[Any, Any]) -> T | None:
        entity = await self.get_entity_by_key(
            key=self.entity_key(cls, pk)
        )
        if entity is None:
            return None
        return self.model_factory(dict(entity), cls)

    async def put(self, obj: pydantic.BaseModel, pk: int | str | ResourceIdentifier[Any, Any]) -> IDatastoreKey:
        key = self.entity_key(type(obj), pk)
        entity = self.entity_factory(key, obj)
        await self.run_in_executor(self.client.put, entity) # type: ignore
        return entity.key # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

import httpx
from aiopki.ext.jose import JOSEObject
from fastapi.security import HTTPBasicCredentials


class IClient(Protocol):
    __module__: str = 'oauthx.lib.protocols'

    @property
    def id(self) -> str:
        ...

    def is_confidential(self) -> bool:
        """Return a boolean indicating if the client is confidential."""
        ...

    async def authenticate(self, credential: HTTPBasicCredentials | JOSEObject | str | None) -> None:
        """Authenticate the client using the given credential.
        """
        ...

    async def userinfo(self, access_token: str, http: httpx.AsyncClient | None = None) -> None:
        """Query the providers' UserInfo endpoint to obtain information
        about the resource owner.
        """
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .useragentexception import UserAgentException


class InvalidAuthorizationResponse(UserAgentException):
    __module__: str = 'oauthx.types'
    template_name: str = 'oauthx/errors/invalid-redirect-parameters.html.j2'
    
    def get_template_names(self) -> list[str] | str:
        return [
            f'oauthx/errors/upstream.{self.error}.html.j2',
            'oauthx/errors/upstream.html.j2',
            self.template_name
        ]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from .basegrant import BaseGrant


class RefreshTokenGrant(BaseGrant):
    grant_type: Literal['refresh_token'] = pydantic.Field(
        default=...,
        description="This value **must** be `refresh_token`."
    )

    refresh_token: str = pydantic.Field(
        default=...,
        description="The refresh token that was issued to the client."
    )

    scope: str | None = pydantic.Field(
        default=None,
        description=(
            "The space-delimited scope of the access request. The requested "
            "scope **must not** include any scope not originally granted by "
            "the resource owner, and if omitted is treated as equal to the "
            "scope originally granted by the resource owner."
        )
    )